"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
from numba import cuda

@cuda.jit(cache=True)   
def add3d_tflite(result,a,b,k_a,k_b,off_a,off_b,activation,out_mul,out_offest):

    x, y,z= cuda.grid(3) 
    
    max_x=result.shape[0]
    max_y=result.shape[1]
    max_z=result.shape[2]

    if ((x >= max_x)  or (y >= max_y) or (z >= max_z)): 
        return 
    
    a1 = (a[x,y,z] - off_a)*k_a
    b1 = (b[x,y,z] - off_b)*k_b
    ris = a1 + b1

    if activation and ris<0:
        ris =0

    result[x,y,z] = round(ris*out_mul + out_offest)


 
@cuda.jit(cache=True) 
def add3dv2_tflite(result,a,b,k_a,k_b,off_a,off_b,activation,out_mul,out_offest):

    x, y= cuda.grid(2) 
    
    max_x=result.shape[0]
    max_y=result.shape[1]
    max_z=result.shape[2]

    if ((x >= max_x)  or (y >= max_y)): 
        return 
    for z in range(max_z):
        a1 = (a[x,y,z] - off_a)*k_a
        b1 = (b[x,y,z] - off_b)*k_b
        ris = a1 + b1

        if activation and ris<0:
            ris =0

        result[x,y,z] = round(ris*out_mul + out_offest)




"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
from numba import cuda, int8


@cuda.jit(cache=True)
def reshape_3d(results, values,size):
    i, j,k = cuda.grid(3)
    
    max_i=values.shape[0]
    max_j=values.shape[1]
    max_k=values.shape[2]

    
    #if the thread coordinates are outside of the image, we ignore the thread:
    if ((i >= max_i)  or (j >= max_j) or (k >= max_k)): 
        return
    
    init_i = i*size[0]
    init_j = j*size[1]
    
    for ii in range(init_i,init_i+size[0]):
        for jj in range(init_j,init_j+size[1]):
            
            results[ii,jj,k] = values[i,j,k]





"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
from numba import cuda, int8,int16,int32,uint8

@cuda.jit(cache=True)
def matmul_tflite(R,A, B,k_mul,k_filtri,k_bias, M,offset1,offset2,activation,out_mul,out_offest,update_bias):

    #MM = cuda.shared.array((128,128),dtype = int16)
    
    #for i in range(128):
    #    for j in range(128):
    #        MM[i,j] = M[i,j]
            
    row, col = cuda.grid(2)
    if row < R.shape[0] and col < R.shape[1]:
        tmp = 0
        ris = k_bias[col] - k_filtri[col] 
        for k in range(A.shape[1]): 
            tmp +=  M[uint8(A[row, k]+offset1) , B[col, k]+offset2]        
            #tmp += A[row, k]*B[col, k]
            
        #ris = k_mul[row]*tmp - k_filtri[col] + k_bias[col]#TODO:in caso di moltiplicazione matriciali(questa di fatto è un vettore per una matrice è)
        ris += k_mul[row]*tmp 
        
        if activation and ris<0:
            ris=0

        R[row, col] = round(ris*out_mul + out_offest)
        
    return   

@cuda.jit(cache=True)
def matmul_tflite_occorenza(R,A, B,k_mul,k_filtri,k_bias, M,offset1,offset2,activation,out_mul,out_offest,occorenza):
    
    row, col = cuda.grid(2)
    if row < R.shape[0] and col < R.shape[1]:
        tmp = 0
        for k in range(A.shape[1]): 
            tmp +=  M[int8(A[row, k])+offset1 , int8(B[col, k]+offset2)]
            occorenza[col,int8(A[row, k])+offset1 , int8(B[col, k]+offset2)] += 1
            #tmp += A[row, k]*B[col, k]
            
        ris = k_mul[row]*tmp - k_filtri[col] + k_bias[col]#TODO:in caso di moltiplicazione matriciali(questa di fatto è un vettore per una matrice è)
        
        if activation and ris<0:
            ris=0

        R[row, col] = round(ris*out_mul + out_offest)
        
    return   

@cuda.jit(cache=True)
def matmul_tflite_mf(R,A, B,k_mul,k_filtri,k_bias, M,offset1,offset2,activation,out_mul,out_offest,idx_multiply):
    
    #MM = cuda.shared.array((128,128),dtype = int16)
    
    #for i in range(128):
    #    for j in range(128):
    #        MM[i,j] = M[i,j]
            
    row, col = cuda.grid(2)
    if row < R.shape[0] and col < R.shape[1]:
        tmp = 0
        for k in range(A.shape[1]):
            
            tmp +=  M[idx_multiply[col],int8(A[row, k])+offset1 , int8(B[col, k]+offset2)]
            
            #tmp += A[row, k]*B[col, k]
            
        ris = k_mul[row]*tmp - k_filtri[col] + k_bias[col]#TODO:in caso di moltiplicazione matriciali(questa di fatto è un vettore per una matrice è)
        
        if activation and ris<0:
            ris=0

        R[row, col] = round(ris*out_mul + out_offest)
        
    return   



TPB = 16
@cuda.jit(cache=True)
def fast_matmul_tflite(R,A, B,k_mul,k_filtri,k_bias, M,offset1,offset2,activation,out_mul,out_offest,update_bias):
    # Define an array in the shared memory
    # The size and type of the arrays must be known at compile time
    sA = cuda.shared.array(shape=(TPB, TPB), dtype=int32)
    sB = cuda.shared.array(shape=(TPB, TPB), dtype=int32)

    x, y = cuda.grid(2)

    tx = cuda.threadIdx.x
    ty = cuda.threadIdx.y
    bpg = cuda.gridDim.x    # blocks per grid

    if x >= R.shape[0] and y >= R.shape[1]:
        # Quit if (x, y) is outside of valid C boundary
        return

    # Each thread computes one element in the result matrix.
    # The dot product is chunked into dot products of TPB-long vectors.
    tmp = 0
    for i in range(bpg):
        # Preload data into shared memory
        sA[ty, tx] = 0
        sB[ty, tx] = 0
        if y < A.shape[0] and (tx + i * TPB) < A.shape[1]:
            sA[ty, tx] = A[y, tx + i * TPB]
        if x < B.shape[1] and (ty + i * TPB) < B.shape[0]:
            sB[ty, tx] = B[ty + i * TPB, x]
        # Wait until all threads finish preloading
        cuda.syncthreads()

        # Computes partial product on the shared memory
        for j in range(TPB):
            #tmp += M[int8(sA[tx, j] )+offset1 , int8(sB[j, ty]+offset2)]   
            tmp += sA[tx, j] * sB[j, ty]

        # Wait until all threads finish computing
        cuda.syncthreads()

    
    ris = k_mul[x]*tmp - k_filtri[y] + k_bias[y] + update_bias[y]
        
    if activation and ris<0:
        ris=0

    R[x, y] = round(ris*out_mul + out_offest)



max_segment = 1
@cuda.jit(cache=True)
def matmul_tflite_fast_pippo(R,A, B,k_mul,k_filtri,k_bias, M,offset1,offset2,activation,out_mul,out_offest,update_bias):

    #MM = cuda.shared.array((128,128),dtype = int16)
    
    #for i in range(128):
    #    for j in range(128):
    #        MM[i,j] = M[i,j]
    
    row , col,segment = cuda.grid(3)

    tmp_shared = cuda.shared.array(shape=(max_segment), dtype=int32)
    if row < R.shape[0] and col < R.shape[1] and segment < max_segment:
        tmp = 0
        #segment = 0
        for k in range(segment*A.shape[1]//max_segment,(1+segment)*A.shape[1]//max_segment): 
            tmp_shared[segment] +=  M[int8(A[row, k])+offset1 , int8(B[col, k]+offset2)]        
            #tmp += A[row, k]*B[col, k]
            
        cuda.syncthreads()
        if segment == 0:
            for i in range(max_segment):
                tmp += tmp_shared[i]
        #ris = k_mul[row]*tmp - k_filtri[col] + k_bias[col]#TODO:in caso di moltiplicazione matriciali(questa di fatto è un vettore per una matrice è)
        
            ris = k_mul[row]*tmp - k_filtri[col] + k_bias[col] + update_bias[col]
        
            if activation and ris<0:
                ris=0
                
                
            R[row, col] =round(ris*out_mul + out_offest)
    return   


import  ctypes,os
from numba import cuda

class Dense_cuda(object):
    def __init__(self, size_H=3, SIZE_W=3, size_filter=3):
        dir_path = os.path.dirname(os.path.abspath(__file__))
        
        self.lib = ctypes.cdll.LoadLibrary(f'{dir_path}/matrixMul.so')
        #self.lib = ctypes.cdll.LoadLibrary('/home/ssaa/Git/inspect-nn/inspectnn/Conv/conv_cuda/convolutionTexture/libconvolutionTexture.so')
        self.size_H = size_H
        self.size_W = SIZE_W
        self.size_filter = size_filter

        # Declare input and output types for each method you intend to use
        #float *d_Dst, int imageW,int imageH, cudaTextureObject_t texSrc
        
        self.lib.setMatrixKernel.argtypes = [ctypes.c_void_p]
        self.lib.setMatrixKernel.restype = ctypes.c_void_p
        
        
        self.lib.MatrixRowsGPU.argtypes = [ctypes.c_void_p]
        self.lib.MatrixRowsGPU.restype =  ctypes.c_void_p
        

        #lib.DataConv
        self.lib.getStruct.argtypes = []
        self.lib.getStruct.restype =  ctypes.c_void_p
        self.obj = self.lib.getStruct()
        #self.obj = lib.convolutionRowsGPU(val)
        
        self.lib.setSizeData.argtypes = [ctypes.c_void_p,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int,ctypes.c_int]
        self.lib.setSizeData.restype =  ctypes.c_void_p
        
        self.lib.setPointerMul.argtypes = [ctypes.c_void_p,ctypes.c_void_p]
        self.lib.setPointerMul.restype = ctypes.c_void_p
        
        
        self.lib.setPointerData.argtypes = [ctypes.c_void_p,ctypes.c_void_p,ctypes.c_void_p,ctypes.c_void_p,
                                                        ctypes.c_void_p,ctypes.c_void_p,ctypes.c_void_p,
                                                        ctypes.c_float,ctypes.c_int]
        self.lib.setPointerData.restype = ctypes.c_void_p
        
        self.lib.print_parameters.argtypes= [ctypes.c_void_p]
        
        self.lib.setPointerInput.argtypes = [ctypes.c_void_p,ctypes.c_void_p,ctypes.c_bool]

    def setPointerInput(self,Input,activation):
        #print("Activation: ",activation)
        self.lib.setPointerInput(self.obj,Input,activation)
        
        
    def setStride(self,x,y,val_null=0):
        self.lib.setStride(self.obj,x,y,val_null)
        
    def setSizeData(self,in_x,in_z,out_x, out_z, k_x, k_y, k_p, size_m):
        self.lib.setSizeData(self.obj,in_x,in_z,out_x, out_z, k_x, k_y, k_p, size_m)
        
        
    def setPointerData(self,Input,d_Dst, Kernel,k_mul,weights,k_bias,out_mul, out_offset):
        self.lib.setPointerData(self.obj,d_Dst, Input, Kernel,k_mul,weights,k_bias,out_mul, out_offset)
            
    def setPointerMul(self, M):
        self.lib.setPointerMul(self.obj,M)
        
        return
    
       
    def MatrixRowsGPU(self):
        cuda.synchronize()
        self.lib.MatrixRowsGPU(self.obj)
        
         
    def setMatrixKernel(self,kernel):
        self.lib.setMatrixKernel(kernel)
    
    def setConvolutiontexSrc(self,ASrc):
        self.lib.setConvolutiontexSrc(ASrc)
        
    def setConvolutiontexSrcv2(self,ASrc):
        self.lib.setConvolutiontexSrcv2(ASrc)
        


    def print_parameters(self):
        self.lib.print_parameters(self.obj)
        
        
    def cuda_set(self,classLayer,A_global_mem ):
        
        #self.test_cuda.setConvolutiontexSrc(A_global_mem.device_ctypes_pointer())
        
        self.setPointerData(A_global_mem.device_ctypes_pointer.value,
                                        classLayer.results.device_ctypes_pointer.value,
                                        classLayer.weights.device_ctypes_pointer.value,
                                        classLayer.k_mul.device_ctypes_pointer.value,
                                        classLayer.k_filtri.device_ctypes_pointer.value,
                                        classLayer.k_bias.device_ctypes_pointer.value,
                                        classLayer.output_mul,classLayer.output_offset)
                
        self.setSizeData(classLayer.input_shape[0],1,classLayer.output_shape[0],1,
                                    classLayer.weights_shape[0],classLayer.weights_shape[1],1,256)
                                    
        self.setPointerInput(A_global_mem.device_ctypes_pointer.value,
                                        classLayer.activation)

        if cuda.is_cuda_array(classLayer.M) :
            self.setPointerMul(classLayer.M.device_ctypes_pointer.value)
           
        
        #self.test_cuda.setStride(self.stride[0],self.stride[1],self.pre_layer.output_offset)


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import numpy as np
from numba import cuda, jit
from inspectnn.BaseLayer import BaseLayer


@cuda.jit(cache=True)   
def concatena_tflite(result,input,offset,size):

    x, y,z= cuda.grid(3) 
    
    max_x=result.shape[0]
    max_y=result.shape[1]
    max_z=size

    if ((x >= max_x)  or (y >= max_y) or (z >= max_z)): 
        return 

    result[x,y,offset+z] = input[x,y,z]
    
#TODO: Agiungre a inspectnn

class Concatenate(BaseLayer):
    def __init__(self, name = "cocatenate"):
        super().__init__(self, name = name)
        self.gpu_layers = None

    def __deepcopy__(self, memo = None):
        return Concatenate(name = self.name)

    def load_weights(self, **kwargs):
        #self.input_shape, self.enable_gpu = kwargs["input_shape"], kwargs["enable_gpu"] 
        self.output_shape = kwargs["output_shape"]
        self.enable_gpu = kwargs["enable_gpu"] 
        
        #self.outputv = np.zeros(self.output_shape)
        if self.enable_gpu:
            self.use_gpu = True
            self.results = cuda.device_array(self.output_shape, dtype=np.int8)
            self.gpu_input_memory = kwargs["gpu_input_memory"]
            #self.pre_layer = kwargs["pre_layer"]
            
            self.output_mul = 1/kwargs["quant_output_k"]
            self.output_offset = kwargs["quant_output_offset"]
        return self.output_shape
    
    #@jit
    def forward_pass(self, **kwargs):
        if self.gpu_layers is None:
            print("concatenate no buoeno! =( ")
            
        else:
            # TODO reshape su GPU
            #cuda.synchronize()
            i = 0
            cuda.synchronize()
            for input in self.gpu_layers:
                #print(input.name)
                #copia 
                concatena_tflite[self.griddim[i], self.blockdim](self.results,input.results,self.offset[i],self.size[i])
                i += 1
               
            #self.results = self.gpu_input_memory.reshape(self.output_shape[0],self.output_shape[1])
            #self.results = self.gpu_input_memory.ravel()
            
            #self.outputv[:,:]=self.results.copy_to_host()
            #print(self.name,"output:\n",self.results.copy_to_host()[0][0])


    def load_input_layer(self,layers,tensors_out):
        
        self.gpu_output_memory = True
        
        self.gpu_layers = []
        self.griddim = []
        self.blockdim = []
        self.offset = [0]
        self.size = []
        
        self.blockdim = (1,1,8)


        for op_in in self.code_tensor_inputs:
        
            #op_in = self.code_tensor_inputs[i]
            #print(op_in)
            self.gpu_layers.append(layers[tensors_out.index(op_in)])
            
            output_shape = self.gpu_layers[-1].output_shape
            self.size.append(output_shape[2])
            self.offset.append(self.offset[-1]+self.size[-1])#incrementa l'offset con il valore della dimensione dell'ultimo tensore
            
            #calcolo della griddim in base al valore di uscita del layer di ingresso (è indispensabile perche vanno lanciati piu kernel in parallelo per ogni tensore di ingresso)
            self.griddim.append((output_shape[0] // self.blockdim[0] + 1, output_shape[1] // self.blockdim[1] + 1,output_shape[2] // self.blockdim[2] + 1))#,n_channels)


            
            

        
        #op_in1=self.code_tensor_inputs[0]
        #op_in2=self.code_tensor_inputs[1]


        

        
      


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""

import numpy as np
from numba import cuda,jit
from inspectnn.BaseLayer import BaseLayer
from inspectnn.PAD.Pad_kernel_tflite import pad_3d


class PadLayer(BaseLayer):
    layer_name = "Pad"
    def __init__(self,type='FakeMinMax', name = "Quantize",quant_nbits = 8):
        super().__init__(self, name = name,quant_nbits=quant_nbits)
        self.type=type
        
        
    def __deepcopy__(self, memo = None):
        return PadLayer( name = self.name,type=self.type,quant_nbits=self.quant_nbits)

    #@jit
    def forward_pass_input(self, inputv):
        if self.gpu_input_memory is None:
            
            if cuda.is_cuda_array(inputv):
                A_global_mem = inputv
            else:
                A_global_mem = cuda.to_device(np.ascontiguousarray(inputv))
            input_mul = 1
            input_offset = 0
        else:
            A_global_mem=self.pre_layer.results
            input_mul = self.pre_layer.output_mul
            input_offset = self.pre_layer.output_offset
        
        A_global_mem = inputv
        #print(self.input_shape ,"-", self.output_shape)
        #print(self.name,"input v1",A_global_mem.copy_to_host()[0])
        
        if True or self.use_3d:
            cuda.synchronize()
            pad_3d[self.griddim, self.blockdim](self.results, A_global_mem, self.output_offset)
        else:
            cuda.synchronize()
            quant_2d[self.griddim, self.blockdim](self.results, A_global_mem,self.output_mul,self.output_offset,input_mul,input_offset)#TODO: mettere input_mul,input_offset in self
        
        #print(self.name,"out",self.results.copy_to_host().shape)
        #print(self.name,"out",self.results.copy_to_host()[4][-2])
        #print(self.name,"avg",np.average(self.results.copy_to_host()))
        if self.gpu_output_memory == False:          
            self.outputv[:,:,:] =self.results.copy_to_host()
            
    
    def forward_pass(self, **kwargs):
        if self.gpu_input_memory is None:
            try:
                inputv = kwargs["inputv"]
            except KeyError as e:
                print(e)
                print(kwargs)
                exit()
            if cuda.is_cuda_array(inputv):
                A_global_mem = inputv
            else:
                A_global_mem = cuda.to_device(np.ascontiguousarray(inputv))
            input_mul = 1
            input_offset = 0
        else:
            A_global_mem=self.pre_layer.results
            input_mul = self.pre_layer.output_mul
            input_offset = self.pre_layer.output_offset

        #print(self.input_shape ,"-", self.output_shape)
        #print(self.name,"input v2",A_global_mem.copy_to_host()[0][0])
        
        if True or self.use_3d:
            cuda.synchronize()
            pad_3d[self.griddim, self.blockdim](self.results, A_global_mem, self.output_offset)
        else:
            cuda.synchronize()
            quant_2d[self.griddim, self.blockdim](self.results, A_global_mem,self.output_mul,self.output_offset,input_mul,input_offset)
        
        #print(self.name,"out",self.results.copy_to_host().shape)
        #print(self.name,"out",self.results.copy_to_host()[0][0])
        #print(self.name,"avg",np.average(self.results.copy_to_host()))
        if self.gpu_output_memory == False:          
            self.outputv[:,:,:] =self.results.copy_to_host()


    def load_weights(self, **kwargs):
        #TODO:mettere tutto a int8
        self.input_shape, self.enable_gpu = kwargs["input_shape"],kwargs["enable_gpu"] 
        self.output_shape = kwargs["output_shape"]      
        self.outputv = np.zeros(self.output_shape,dtype=np.int16)#TODO: mettere dentro if parametrico

        self.dimension = len(self.input_shape)
        if self.enable_gpu:
            self.use_gpu=True

            if self.dimension == 3 and self.input_shape[2] > 1:
                self.use_3d = True
                self.blockdim = (2,2,2)
                self.griddim = (self.output_shape[0] // self.blockdim[0] + 1, self.output_shape[1] // self.blockdim[1] + 1,self.output_shape[2] // self.blockdim[2] + 1)#,n_channels)
            elif self.dimension == 2 or self.input_shape[2] == 1:
                self.use_3d = False
                self.blockdim = (4,4)
                self.griddim = (self.output_shape[0] // self.blockdim[0] + 1, self.output_shape[1] // self.blockdim[1] + 1)#,n_channels)
            else:
                print("Errore Input shape:",self.input_shape)


            self.results = cuda.device_array(self.output_shape, dtype=np.int8)#int16 o 8
            
            self.gpu_input_memory = kwargs["gpu_input_memory"]
            self.output_mul= 1/kwargs["quant_output_k"]
            self.output_offset= kwargs["quant_output_offset"]

        return self.output_shape
    


import time

import numpy as np

from inspectnn.Conv.ConvLayer_TFLITE import ConvLayer_TFLITE
from inspectnn.Dense.DenseLayer_TFLITE import DenseLayer_TFLITE
from inspectnn.Pooling.PoolingLayer_TFLITE import PoolingLayer_TFLITE
from inspectnn.Mean.MeanLayer_TFLITE import MeanLayer_TFLITE
from inspectnn.Flatten.FlattenLayer import FlattenLayer
from inspectnn.PAD.PadLayer import PadLayer
from inspectnn.Add.AddLayer import AddLayer

from inspectnn.Model.NetworkModel_tflite import NetworkModelTflite
from inspectnn.Model.BaseModel_tflite import BaseModel_tflite
from inspectnn.Quantizate.QuantizateLayer import QuantizateLayer
from inspectnn.Concatenate.ConcatenateLayer import Concatenate
from inspectnn.SoftmaxLayer import SoftmaxLayer
from inspectnn.Reshape.ReshapeLayer import ReshapeLayer

import tflite
import numpy as np, time, tensorflow as tf

class GenericModelTflite(BaseModel_tflite):
    def __init__(self,path_tf_file, input_no_normalizate=True):
        
        print('\033[92m#######################\033[0m')
        print('\033[92m####   InspectNN   ####\033[0m')
        print('\033[92m#######################\033[0m\n')
        
        self.quant_nbits = 7
        save_csv_path=''
        self.input_no_normalizate = input_no_normalizate
 
        #print("path file tflite:",path_tf_file)
        self.interpreter = tf.lite.Interpreter(model_path=path_tf_file,experimental_preserve_all_tensors=True)
        self.interpreter.allocate_tensors()
        #apri il file tflite
        f = open(path_tf_file,"rb")
        buf = f.read()

        self.model = tflite.Model.GetRootAsModel(buf,0)

        #print("len:",self.model.SubgraphsLength())
        graf = self.model.Subgraphs(0)
        input_shape = self.interpreter.get_tensor_details()[graf.Operators(0).Inputs(0)]['shape'][1:]
        super().__init__(None,self.quant_nbits,input_shape,save_csv_path)
       
        #ssuper().__init__(None,self.quant_nbits,input_shape,save_csv_path)
        layers = []
        self.data_layer = []
        self.data_layer_quant = []
        self.learned_parameters = []

        self.n_moltiplicazioni = []
        self.tensor_output_layer = []

        
        #Crea i singoli layer
        for i in range(graf.OperatorsLength()-1):#ELIMINA L'ULTIMO LAYER CHE È QUELLO DI OUTPUT
            op=graf.Operators(i)      
            op_code = self.model.OperatorCodes(op.OpcodeIndex())     
            #print(i,tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())],op.InputsLength(),op.OutputsLength())
            #for j in range(op.InputsLength()):
            #    print("\t",op.Inputs(j))
           
            layers.append(self.__ImportLayer__(op))
            
            layers[-1].id_output_layer = op.Outputs(0) #Layers id tflite
            self.tensor_output_layer.append(op.Outputs(0))
            #print(layers[-1].name,op.Outputs(0),op.Inputs(0))
            #print("OUT:",op.Outputs(0))

        #collega le uscite dei layer
        layers[0].gpu_output_memory = True
        for i in range(1,graf.OperatorsLength()-2):
            layers[i].load_input_layer(layers,self.tensor_output_layer)
        
        layers[-2].end_layer =True
        layers[-2].gpu_output_memory = False
        layers[-1].pre_layer = layers[-2]

        self.net =  NetworkModelTflite(quant_data_scale=self.data_layer_quant,
                                       quant_data=self.data_layer, 
                                       learned_parameters = self.learned_parameters, 
                                       input_shape = input_shape,
                                       layers = layers)

        #print(path_tf_file,np.sum(self.n_moltiplicazioni))

        #TODO: da controlare, dovrebbe far risparmiare un po di memoria eprche non vengono piu usati
        #del self.interpreter, self.model
        #del self.tensor_output_layer
        
        #TODO: far si che i risultati rimangono su gpu e solo alla fine di tutte le valutazioni si caricano su cpu
        
    def __LayerOpcodeConverter__(self,id):
        return self.model.OperatorCodes(id).BuiltinCode() 

    def __ReadGenericParameters__(self,op,name_id=0):
        #print(self.interpreter.get_tensor_details()[op.Outputs(0)]['name'].split(';')[-name_id])
        try:
            name_layer=self.interpreter.get_tensor_details()[op.Outputs(0)]['name'].split(';')[name_id].split('/')[1]
        except:
            name_layer=self.interpreter.get_tensor_details()[op.Outputs(0)]['name']
            
        input_shape = self.interpreter.get_tensor_details()[op.Inputs(0)]['shape'][1:]

        return name_layer,input_shape
    
    def __ImportLayer__(self,op):
        disponibilita_gpu=True
        #print(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())])
        #TODO: parametrizare meglio i layer prendendo i dati da op
        layer=None
        op_opt = op.BuiltinOptions()
        
        if(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())] == "QUANTIZE"):
            
            #quantize_op = tflite.QuantizeOptions()
            #quantize_op.Init(op_opt)
            #print(quantize_op)
            
            name_layer,input_shape = self.__ReadGenericParameters__(op)
            layer = QuantizateLayer(quant_nbits =  self.quant_nbits, name = 'Q_' + name_layer)
            layer.code_tensor_inputs = [op.Inputs(i) for i in range(1)]

            if op.Inputs(0) == 0 and self.input_no_normalizate:
                quant_output_k = 1#self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][0]#*255
            else:
                quant_output_k = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][0]

            quant_output_offset = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][1]

            layer.load_weights(input_shape = input_shape,enable_gpu=disponibilita_gpu,gpu_input_memory=None,
                               quant_output_k=quant_output_k,quant_output_offset=quant_output_offset)
                
        elif(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())] == "CONV_2D"):
            
            conv_op = tflite.Conv2DOptions()
            conv_op.Init(op_opt.Bytes, op_opt.Pos)

            name_layer,input_shape = self.__ReadGenericParameters__(op,-1)#o 1 dipende =()
            type_attivation=self.interpreter.get_tensor_details()[op.Outputs(0)]['name'].split(';')[0].split('/')[-1]

            if conv_op.Padding() == tflite.Padding.SAME :
                padding_type ='same'
            else:
                padding_type ='valid'
            
            layer = ConvLayer_TFLITE(stride = (conv_op.StrideW(), conv_op.StrideH()), padding = (0, 0), activation = type_attivation, padding_type = padding_type,
                                      quant_nbits =  self.quant_nbits, name = name_layer,multiplier=self.exact_multiplier)#multiplier=MUL_2bit_importata
           
            self.learned_parameters.append(self.interpreter.get_tensor(op.Inputs(1)))#pesi weights
            self.learned_parameters.append(self.interpreter.get_tensor(op.Inputs(2)))#pesi bias
            
            quantization_pesi = self.interpreter.get_tensor_details()[op.Inputs(1)]['quantization_parameters']['scales']
            quantization_bias = self.interpreter.get_tensor_details()[op.Inputs(2)]['quantization_parameters']['scales']

            quant_input_k = self.interpreter.get_tensor_details()[op.Inputs(0)]['quantization'][0]
            quant_input_offset =  self.interpreter.get_tensor_details()[op.Inputs(0)]['quantization'][1]

            output_shape=self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:]
            
            quant_output_k = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][0]
            quant_output_offset = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][1]
            
            layer.code_tensor_inputs = [op.Inputs(i) for i in range(1)]
            layer.load_weights(input_shape = input_shape,output_shape=output_shape, weights = self.learned_parameters[-2], biases = self.learned_parameters[-1],
                                                                        enable_gpu=disponibilita_gpu,gpu_input_memory=None,
                                                                        quant_data=quantization_pesi,quant_bias=quantization_bias,
                                                                        quant_input_k=quant_input_k,quant_input_offset=quant_input_offset,
                                                                        quant_output_k=quant_output_k,quant_output_offset=quant_output_offset
            
                                                                        )
            #dimensione kernel(senza il numero di filtri, presenti nel output) * dimensione output
            #layer.kernel_shape = np.prod(self.interpreter.get_tensor(op.Inputs(1)).shape[1:])
            #print(f"Nome Layer: {name_layer} Termine1:", self.interpreter.get_tensor(op.Inputs(1)).shape[1:], "Termine2:", output_shape)
            #print("input_shape: ", input_shape, "output_shape: ", output_shape)
            layer.n_moltiplicazioni=np.prod(self.interpreter.get_tensor(op.Inputs(1)).shape[1:])*np.prod(output_shape)
            #print(layer.n_moltiplicazioni)
           
        elif(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())] == "DEPTHWISE_CONV_2D"):
            
            conv_op = tflite.Conv2DOptions()
            conv_op.Init(op_opt.Bytes, op_opt.Pos)
            
            stride =  [conv_op.StrideW(), conv_op.StrideH()] #self.interpreter.get_tensor_details()[op.Inputs(0)]['shape'][1:3]//self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:3]
            #pool_size = [conv_op.FilterWidth(), conv_op.FilterHeight()]#stride
            #print("ADD Layer: DEPTHWISE_CONV_2D")
            name_layer,input_shape = self.__ReadGenericParameters__(op,0)#o 1 dipende =()
            type_attivation=self.interpreter.get_tensor_details()[op.Outputs(0)]['name'].split(';')[0].split('/')[-1]

            if conv_op.Padding() == tflite.Padding.SAME :
                padding_type ='same'
            else:
                padding_type ='valid'
            
            layer = ConvLayer_TFLITE(stride = (conv_op.StrideW(), conv_op.StrideH()), padding = (0, 0), activation = type_attivation, padding_type = padding_type,
                                      quant_nbits =  self.quant_nbits, name = name_layer,multiplier=self.exact_multiplier,DEPTHWISE_CONV_2D=True)#multiplier=MUL_2bit_importata
           
            self.learned_parameters.append(self.interpreter.get_tensor(op.Inputs(1)))#pesi weights
            self.learned_parameters.append(self.interpreter.get_tensor(op.Inputs(2)))#pesi bias
            
            quantization_pesi = self.interpreter.get_tensor_details()[op.Inputs(1)]['quantization_parameters']['scales']
            quantization_bias = self.interpreter.get_tensor_details()[op.Inputs(2)]['quantization_parameters']['scales']

            quant_input_k = self.interpreter.get_tensor_details()[op.Inputs(0)]['quantization'][0]
            quant_input_offset =  self.interpreter.get_tensor_details()[op.Inputs(0)]['quantization'][1]

            output_shape=self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:]
            
            quant_output_k = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][0]
            quant_output_offset = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][1]
            
            layer.code_tensor_inputs = [op.Inputs(i) for i in range(1)]
            layer.load_weights(input_shape = input_shape,output_shape=output_shape, weights = self.learned_parameters[-2], biases = self.learned_parameters[-1],
                                                                        enable_gpu=disponibilita_gpu,gpu_input_memory=None,
                                                                        quant_data=quantization_pesi,quant_bias=quantization_bias,
                                                                        quant_input_k=quant_input_k,quant_input_offset=quant_input_offset,
                                                                        quant_output_k=quant_output_k,quant_output_offset=quant_output_offset
            
                                                                        )
            #dimensione kernel(senza il numero di filtri, presenti nel output) * dimensione output
            #layer.kernel_shape = np.prod(self.interpreter.get_tensor(op.Inputs(1)).shape[1:])
            #print(f"Nome Layer: {name_layer} Termine1:", self.interpreter.get_tensor(op.Inputs(1)).shape[1:], "Termine2:", output_shape)
            #print("input_shape: ", input_shape, "output_shape: ", output_shape)
            layer.n_moltiplicazioni=np.prod(self.interpreter.get_tensor(op.Inputs(1)).shape[:-1])*np.prod(output_shape)
            #print(layer.n_moltiplicazioni)
            
                  
        elif(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())] == "FULLY_CONNECTED"):

            type_attivation=self.interpreter.get_tensor_details()[op.Outputs(0)]['name'].split(';')[1].split('/')[-1]
            name_layer,input_shape = self.__ReadGenericParameters__(op,1)
            

            if ( type_attivation != 'relu' and type_attivation != 'Relu'):
                type_attivation='softmax'
            layer=DenseLayer_TFLITE(activation = type_attivation, quant_nbits =  self.quant_nbits, name = name_layer,multiplier=self.exact_multiplier)
            self.learned_parameters.append(self.interpreter.get_tensor(op.Inputs(1)))#pesi weights

            self.learned_parameters.append(self.interpreter.get_tensor(op.Inputs(2)))#pesi bias
            
            quantization_pesi = self.interpreter.get_tensor_details()[op.Inputs(1)]['quantization_parameters']['scales']
            quantization_bias = self.interpreter.get_tensor_details()[op.Inputs(2)]['quantization_parameters']['scales']

            quant_input_k = self.interpreter.get_tensor_details()[op.Inputs(0)]['quantization'][0]
            quant_input_offset =  self.interpreter.get_tensor_details()[op.Inputs(0)]['quantization'][1]

            quant_output_k = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][0]
            quant_output_offset = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][1]
            output_shape=self.interpreter.get_tensor_details()[op.Outputs(0)]['shape']

            layer.code_tensor_inputs = [op.Inputs(i) for i in range(1)]

            layer.load_weights(input_shape = input_shape, output_shape=output_shape,weights = self.learned_parameters[-2], biases = self.learned_parameters[-1],
                                                                        enable_gpu=disponibilita_gpu,gpu_input_memory=None,
                                                                        quant_data=quantization_pesi,quant_bias=quantization_bias,
                                                                        quant_input_k=quant_input_k,quant_input_offset=quant_input_offset,
                                                                        quant_output_k=quant_output_k,quant_output_offset=quant_output_offset
                                                                        )
            #dimensione input * output
            print("input_shape: ", input_shape, "output_shape: ", output_shape)
            layer.n_moltiplicazioni=np.prod(output_shape)*np.prod(input_shape)
        
        #TODO: unire i due pool cone la variaible letta da pool_op
        elif(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())] == "AVERAGE_POOL_2D"):
            name_layer,input_shape = self.__ReadGenericParameters__(op)

            pool_op = tflite.Pool2DOptions()
            pool_op.Init(op_opt.Bytes, op_opt.Pos)
            
            stride =  [pool_op.StrideW(), pool_op.StrideH()] #self.interpreter.get_tensor_details()[op.Inputs(0)]['shape'][1:3]//self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:3]
            pool_size = [pool_op.FilterWidth(), pool_op.FilterHeight()]#stride
            
            output_shape=self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:]
            layer = PoolingLayer_TFLITE(stride = stride, pool_size = pool_size, pooling = "avg", name = name_layer)#TODO vare averge pool
            quant_output_k = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][0]
            quant_output_offset = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][1]

            layer.code_tensor_inputs = [op.Inputs(i) for i in range(1)]
            layer.load_weights(input_shape = input_shape,output_shape=output_shape,enable_gpu=disponibilita_gpu,gpu_input_memory=None,
                               quant_output_k=quant_output_k,quant_output_offset=quant_output_offset)

        elif(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())] == "MAX_POOL_2D"):
            name_layer,input_shape = self.__ReadGenericParameters__(op)
            
            pool_op = tflite.Pool2DOptions()
            pool_op.Init(op_opt.Bytes, op_opt.Pos)
            
            stride =  [pool_op.StrideW(), pool_op.StrideH()]#self.interpreter.get_tensor_details()[op.Inputs(0)]['shape'][1:3]//self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:3]
            pool_size = [pool_op.FilterWidth(), pool_op.FilterHeight()]#stride
            output_shape=self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:]
            
            layer = PoolingLayer_TFLITE(stride = stride, pool_size =pool_size, pooling = "max", name = name_layer)

            quant_output_k = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][0]
            quant_output_offset = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][1]

            layer.code_tensor_inputs = [op.Inputs(i) for i in range(1)]
            layer.load_weights(input_shape = input_shape,output_shape = output_shape, enable_gpu=disponibilita_gpu,gpu_input_memory=None,
                               quant_output_k=quant_output_k,quant_output_offset=quant_output_offset)
        
        elif(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())] == "RESHAPE"):
            name_layer=self.interpreter.get_tensor_details()[op.Outputs(0)]['name'].split(';')[0].split('/')[1]
            input_shape = self.interpreter.get_tensor_details()[op.Inputs(0)]['shape'][1:]
            layer = FlattenLayer(name = name_layer)

            layer.code_tensor_inputs = [op.Inputs(i) for i in range(1)]
            layer.load_weights(input_shape = input_shape,enable_gpu=disponibilita_gpu,gpu_input_memory=None)

        elif(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())] == "ADD"):
            type_attivation=self.interpreter.get_tensor_details()[op.Outputs(0)]['name'].split(';')[0].split('/')[-1]
            name_layer=self.interpreter.get_tensor_details()[op.Outputs(0)]['name'].split(';')[1].split('/')[-2]

            layer = AddLayer(name = name_layer, activation = type_attivation)

            input_shape_A = self.interpreter.get_tensor_details()[op.Inputs(0)]['shape']
            input_shape_B = self.interpreter.get_tensor_details()[op.Inputs(1)]['shape']
            output_shape=self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:]

            quant_output_k = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][0]
            quant_output_offset = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][1]

            layer.code_tensor_inputs = [op.Inputs(i) for i in range(2)]
            layer.load_weights(input_shape_A = input_shape_A,input_shape_B=input_shape_B,output_shape=output_shape,
                               quant_output_k=quant_output_k,quant_output_offset=quant_output_offset,
                               enable_gpu=disponibilita_gpu,gpu_input_memory=None)

        elif(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())] == "SOFTMAX"):
            type_attivation=self.interpreter.get_tensor_details()[op.Outputs(0)]['name'].split(';')[0].split('/')[-1]
            name_layer=self.interpreter.get_tensor_details()[op.Outputs(0)]['name'].split('/')[-1]

            output_shape=self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:]

            quant_output_k = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][0]
            quant_output_offset = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][1]
            input_shape = self.interpreter.get_tensor_details()[op.Inputs(0)]['shape']

            layer = SoftmaxLayer(name = name_layer, activation = type_attivation)

            layer.load_weights(input_shape = input_shape,output_shape=output_shape,
                               quant_output_k=quant_output_k,quant_output_offset=quant_output_offset)

        elif(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())] == "PAD"):
    
            name_layer=self.interpreter.get_tensor_details()[op.Outputs(0)]['name'].split(';')[0].split('/')[1]
            input_shape = self.interpreter.get_tensor_details()[op.Inputs(0)]['shape'][1:]
            output_shape=self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:]
            padding_shape = self.interpreter.get_tensor_details()[op.Inputs(1)]['shape'][1:]
            
            quant_output_k = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][0]
            quant_output_offset = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][1]
            
            layer = PadLayer(name = name_layer)

            layer.code_tensor_inputs = [op.Inputs(i) for i in range(1)]
            layer.load_weights(input_shape = input_shape,output_shape=output_shape,enable_gpu=disponibilita_gpu,gpu_input_memory=None,
                               quant_output_k=quant_output_k,quant_output_offset=quant_output_offset
                               )
            
        #TODO: sostituire con un layer Mean piu ottimizato (ne vale la pena?)
        elif(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())] == "MEAN"):
            name_layer,input_shape = self.__ReadGenericParameters__(op)
            

            input_shape = self.interpreter.get_tensor_details()[op.Inputs(0)]['shape'][1:]
            output_shape=self.interpreter.get_tensor_details()[op.Outputs(0)]['shape']
             
            stride =  self.interpreter.get_tensor_details()[op.Inputs(0)]['shape'][1:3]//self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:3]
            pool_size = input_shape
            
            #layer = MeanLayer_TFLITE
            layer = PoolingLayer_TFLITE(stride = stride, pool_size = pool_size, pooling = "avg", name = name_layer)#TODO vare averge pool
            quant_output_k = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][0]
            quant_output_offset = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][1]

            layer.code_tensor_inputs = [op.Inputs(i) for i in range(1)]
            layer.load_weights(input_shape = input_shape,output_shape=output_shape,enable_gpu=disponibilita_gpu,gpu_input_memory=None,
                               quant_output_k=quant_output_k,quant_output_offset=quant_output_offset)
        
        elif(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())] == "REDUCE_MAX"):
            name_layer,input_shape = self.__ReadGenericParameters__(op)
            

            input_shape = self.interpreter.get_tensor_details()[op.Inputs(0)]['shape'][1:]
            output_shape=self.interpreter.get_tensor_details()[op.Outputs(0)]['shape']
             
            stride =  self.interpreter.get_tensor_details()[op.Inputs(0)]['shape'][1:3]//self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:3]
            pool_size = input_shape
            
            #layer = MeanLayer_TFLITE
            layer = PoolingLayer_TFLITE(stride = stride, pool_size = pool_size, pooling = "avg", name = name_layer)#TODO vare averge pool
            quant_output_k = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][0]
            quant_output_offset = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][1]

            layer.code_tensor_inputs = [op.Inputs(i) for i in range(1)]
            layer.load_weights(input_shape = input_shape,output_shape=output_shape,enable_gpu=disponibilita_gpu,gpu_input_memory=None,
                               quant_output_k=quant_output_k,quant_output_offset=quant_output_offset)  
        elif(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())] == "RESIZE_NEAREST_NEIGHBOR"):
            name_layer,input_shape = self.__ReadGenericParameters__(op)
            #stride =  [7, 7] #self.interpreter.get_tensor_details()[op.Inputs(0)]['shape'][1:3]//self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:3]
            pool_size = input_shape
               
            input_shape = self.interpreter.get_tensor_details()[op.Inputs(0)]['shape'][1:]
            output_shape=self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:]
            
            
            quant_output_k = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][0]
            quant_output_offset = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][1]

            stride = (output_shape/input_shape)[0:2].astype(int)
            #layer = MeanLayer_TFLITE
            layer = ReshapeLayer(name = name_layer,dim_size = stride)
            
            layer.load_weights(input_shape = input_shape,output_shape=output_shape, enable_gpu=disponibilita_gpu,gpu_input_memory=None,
                               quant_output_k=quant_output_k,quant_output_offset=quant_output_offset)
            
            layer.code_tensor_inputs = [op.Inputs(i) for i in range(1)]
            
            
        elif(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())] == "CONCATENATION"):
            #print(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())],"Fail supported ")
            
            type_attivation=self.interpreter.get_tensor_details()[op.Outputs(0)]['name'].split(';')[0].split('/')[-1]
            name_layer=self.interpreter.get_tensor_details()[op.Outputs(0)]['name'].split(';')[0].split('/')[-1]

            layer = Concatenate(name = name_layer)
            
            output_shape=self.interpreter.get_tensor_details()[op.Outputs(0)]['shape'][1:]

            quant_output_k = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][0]
            quant_output_offset = self.interpreter.get_tensor_details()[op.Outputs(0)]['quantization'][1]

            layer.code_tensor_inputs = [op.Inputs(i) for i in range(op.InputsLength())]
            layer.load_weights(output_shape=output_shape,quant_output_k=quant_output_k,
                               quant_output_offset=quant_output_offset,
                               enable_gpu=disponibilita_gpu,gpu_input_memory=None)
            
               
            
        else:
            print(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())],"Not supported")
            
            exit(tflite.utils.BUILTIN_OPCODE2NAME[self.__LayerOpcodeConverter__(op.OpcodeIndex())]+" Not supported")
        return layer
    
  


from sklearn.model_selection import train_test_split
import numpy as np, tensorflow as tf, random

def Dataset_mnist(num_of_samples,sequenzial=False,PATH=None):
    
    (x_train, y_train), (x_test, y_test) = tf.keras.datasets.mnist.load_data()
    x_train = x_train.reshape(x_train.shape[0], 28, 28, 1)
    x_test = x_test.reshape(x_test.shape[0], 28, 28, 1)
    
    x_train, x_val, y_train, y_val = train_test_split(x_train, y_train, test_size=0.1, random_state=1)


    # Preprocess the data
    x_train_a = x_train.astype('float32')
    x_test_a = x_test.astype('float32')
    x_val_a = x_val.astype('float32')

    x_train_a /= 255
    x_test_a /= 255
    x_val_a /= 255

    x_test = np.uint8(x_test)
    if(num_of_samples > 0):
        if sequenzial :
            dataset = range( num_of_samples)
        else:
            dataset = random.sample(range(len(x_test_a)), num_of_samples)
    else:
        dataset = [-num_of_samples]

    if(num_of_samples == 1):
        print("Dataset:",dataset)
    #print(x_test.shape[1:])

    c10_images = [x_test_a[i] for i in dataset ]
    c10_labels = [y_test[i] for i in dataset]


    c10_images_apx = [x_test[i] for i in dataset ]
    c10_labels_apx = c10_labels


    return c10_images_apx,c10_labels_apx,c10_images,c10_labels


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
from numba import cuda, int8,int32,uint8
import numpy as np
from tqdm import tqdm

@cuda.jit(cache=True)    
def convolve_tflite_occorenza(result,a,b,params,k_mul,k_filtri,k_bias,M,activation,out_mul,out_offest,stride,occorenza,K_p):

    x, y,p = cuda.grid(3) 
    n_channels = params[0]
    offset1 = params[1]
    offset2 = params[2]
    p=K_p*p
    max_x=result.shape[0]
    max_y=result.shape[1]

    if ((x >= max_x)  or (y >= max_y) or (p >=n_channels)): 
        return  


    xi = stride[0] * x
    yi = stride[1] * y

    range_i = range(xi,xi+b.shape[1])
    range_j = range(yi,yi+b.shape[2])
    range_p = range(0,K_p)
    
    s = cuda.local.array(shape=(K_p), dtype=int32)

    for pp in range_p:
        s[pp]=0
        #calcolo cella convoluzione
        #todo: aggiungere lo step
    for i in range_i:
        for j in range_j:
            for k in range(b.shape[3]):
                #s+=1
                A = uint8(a[i,j,k]+offset1)
                for pp in range_p:
                    s[pp] += M[A,uint8(b[pp+p,i-xi,j-yi,k]+offset2)]
                    occorenza[pp+p,A,int8(b[pp+p,i-xi,j-yi,k]+offset2)] += 1

    for pp in range_p:
        ris = s[pp]*k_mul[pp+p] - k_filtri[pp+p] + k_bias[pp+p]

        if activation and ris<0:
            ris =0
        result[x,y,pp+p] = round(ris*out_mul + out_offest)


@cuda.jit(cache=True)    
def convolve_tflite(result,a,b,params,k_mul,k_filtri,k_bias,
                    M,activation,out_mul,out_offest,stride,update_bias,K_p):

    x, y,p = cuda.grid(3) 
    n_channels = params[0]
    offset1 = params[1]
    offset2 = params[2]
    p=K_p*p
    max_x=result.shape[0]
    max_y=result.shape[1]

    if ((x >= max_x)  or (y >= max_y) or (p >=n_channels)): 
        return  


    xi = stride[0] * x
    yi = stride[1] * y

    range_i = range(xi,xi+b.shape[1])
    range_j = range(yi,yi+b.shape[2])
    range_p = range(0,K_p)
    
    s = cuda.local.array(shape=(16), dtype=int32)

    for pp in range_p:
        s[pp]=0
        #calcolo cella convoluzione
        #todo: aggiungere lo step
        
    for i in range_i:
        for j in range_j:
            for k in range(b.shape[3]):
                #s+=1
                A = uint8(a[i,j,k]+offset1)
                for pp in range_p:
                    s[pp] += M[A,uint8(b[pp+p,i-xi,j-yi,k]+offset2)]

    for pp in range_p:
        ris = s[pp]*k_mul[pp+p] - k_filtri[pp+p] + k_bias[pp+p] 

        if activation and ris<0:
            ris =0
        if activation == 6 and ris>6:
            ris =6
        result[x,y,pp+p] = round(ris*out_mul + out_offest)

@cuda.jit(cache=True)  
def convolve_tflite_padding(result, a,b,params,k_mul,k_filtri,k_bias,M,activation,out_mul,out_offest,value_null,stride,update_bias,K_p):

    x, y,p = cuda.grid(3)
    n_channels = params[0]
    offset1 = params[1]
    offset2 = params[2]  
    max_x=result.shape[0]
    max_y=result.shape[1]

    p=K_p*p
    if ((x >= max_x)  or (y >= max_y) or (p >=n_channels)): 
        return
   
    xi = stride[0] * x
    yi = stride[1] * y
    
    delta_i = int((b.shape[1]-1)/2)
    delta_j = int((b.shape[2]-1)/2)
    

    if(stride[0] > 1):
        offset_i=xi -delta_i+1 
        
    else:
        offset_i=xi-delta_i
        
       
        

    if(stride[1] > 1):
        offset_j=yi-delta_j+1
       
    else:
        offset_j=yi-delta_j
    
    delta_i=2*delta_i
    delta_j=2*delta_j

    range_i = range(offset_i,offset_i+delta_i+1)
    range_j = range(offset_j,offset_j+delta_j+1)
    range_p = range(0,K_p)
    
        #calcolo cella convoluzione
        #todo: aggiungere lo step
    s = cuda.local.array(shape=(16), dtype=int32)

    for pp in range_p:
        s[pp]=0

    for i in range_i:
        for j in range_j:
            for k in range(b.shape[3]):
                #s+=1
                if i >= 0 and  i<a.shape[0] and j >= 0 and j < a.shape[1]:
                    A=uint8(a[i,j,k]+offset1)
                    for pp in range_p:
                        s[pp] += M[A,uint8(b[pp+p,i-offset_i,j-offset_j,k]+offset2)]
                else:
                    A=uint8(value_null+offset1)
                    for pp in range_p:
                        s[pp] += M[A,uint8(b[pp+p,i-offset_i,j-offset_j,k]+offset2)]

    for pp in range_p:
        ris = s[pp]*k_mul[pp+p] - k_filtri[pp+p] + k_bias[pp+p]

        if activation and ris<0:
            ris =0
        
        if activation == 6 and ris>6:
            ris =6
            
        result[x,y,pp+p] = round(ris*out_mul + out_offest)


@cuda.jit(cache=True)  
def convolve_deepwise(result, a,b,params,k_mul,k_filtri,k_bias,M,activation,out_mul,out_offest,stride,update_bias):

    x, y,p = cuda.grid(3)
    n_channels = result.shape[2]
    offset1 = params[1]
    offset2 = params[2]  
    max_x=result.shape[0]
    max_y=result.shape[1]

   
    if ((x >= max_x)  or (y >= max_y) or (p >=n_channels)): 
        return
   
    xi = stride[0] * x
    yi = stride[1] * y

    range_i = range(xi,xi+b.shape[1])
    range_j = range(yi,yi+b.shape[2])
    
    
        #calcolo cella convoluzione
        #todo: aggiungere lo step

    s = cuda.local.array(shape=(1), dtype=int32)

    k=p
    for i in range_i:
        for j in range_j:
            
            #s+=1
            A = uint8(a[i,j,k]+offset1)
            s[0] += M[A,uint8(b[0,i-xi,j-yi,k]+offset2)]

    
    ris = s[0]*k_mul[p] - k_filtri[p] + k_bias[p] 

    if activation>0 and ris<0:
        ris =0
        
    if activation == 6 and ris>6: ris =6
    result[x,y,p] = round(ris*out_mul + out_offest)


@cuda.jit(cache=True)  
def convolve_deepwise_padding(result, a,b,params,k_mul,k_filtri,k_bias,M,activation,out_mul,out_offest,value_null,stride,update_bias):

    x, y,p = cuda.grid(3)
    n_channels = result.shape[2]
    offset1 = params[1]
    offset2 = params[2]  
    max_x=result.shape[0]
    max_y=result.shape[1]

   
    if ((x >= max_x)  or (y >= max_y) or (p >=n_channels)): 
        return
   
    xi = stride[0] * x
    yi = stride[1] * y
    
    delta_i = int((b.shape[1]-1)/2)
    delta_j = int((b.shape[2]-1)/2)
    

    if(stride[0] > 1):
        offset_i=xi
        delta_i=2*delta_i
    else:
        offset_i=xi-delta_i
        

    if(stride[1] > 1):
        offset_j=yi
        delta_j=2*delta_j
    else:
        offset_j=yi-delta_j

    range_i = range(offset_i,xi+delta_i+1)
    range_j = range(offset_j,yi+delta_j+1)
    
    
        #calcolo cella convoluzione
        #todo: aggiungere lo step

    s = cuda.local.array(shape=(1), dtype=int32)

    k=p
    for i in range_i:
        for j in range_j:
        
            #s+=1
            if i >= 0 and  i<a.shape[0] and j >= 0 and j < a.shape[1]:
                A=uint8(a[i,j,k]+offset1)
                
                s[0] += M[A,uint8(b[0,i-offset_i,j-offset_j,p]+offset2)]
            else:
                A=uint8(value_null+offset1)
                
                s[0] += M[A,uint8(b[0,i-offset_i,j-offset_j,p]+offset2)]

    
    ris = s[0]*k_mul[p] - k_filtri[p] + k_bias[p] 

    if activation > 0 and ris<0:
        ris =0
    
        
    if activation == 6 and ris>6: ris =6
    
    result[x,y,p] = round(ris*out_mul + out_offest)

@cuda.jit(cache=True)  
def convolve_tflite_fake(result, a,b,params,k_mul,k_filtri,k_bias,M,activation,out_mul,out_offest,stride,update_bias,K_p):

    #impone che la dimensione del kernel sia 1x1
    x, y,p = cuda.grid(3)    
    max_x=result.shape[0]
    max_y=result.shape[1]
    
    p=K_p*p
    n_channels = params[0]
    if ((x >= max_x)  or (y >= max_y) or (p >=n_channels)): 
        return
    
    offset1 = params[1]
    offset2 = params[2]
    
    xi = stride[0] * x
    yi = stride[1] * y
    
    range_p = range(0,K_p)
    
        #calcolo cella convoluzione
        #todo: aggiungere lo step
    s = cuda.local.array(shape=(16), dtype=int32)

    for pp in range_p:
        s[pp]=0

    for k in range(b.shape[3]):

        A=uint8(a[xi,yi,k]+offset1)
        for pp in range_p:
            #s[pp] += M[A,uint8(b[pp+p,0,0,k]+offset2)]
            s[pp] += A*b[pp+p,0,0,k]

    for pp in range_p:
        
        ris = s[pp]*k_mul[pp+p] - k_filtri[pp+p] + k_bias[pp+p] # #unire k_filtri e k_bias

        if activation and ris<0:
            ris =0
            
        if activation == 6 and ris>6:
            ris =6
            
        result[x,y,pp+p] = round(ris*out_mul + out_offest)


@cuda.jit(cache=True)
def quant_3d_FakeMinMax(results, values,tmp_max,tmp_min,quant_nbits):

    i, j,k = cuda.grid(3)
    
    max_i=results.shape[0]
    max_j=results.shape[1]
    max_k=results.shape[2]
    
    #if the thread coordinates are outside of the image, we ignore the thread:
    if ((i >= max_i)  or (j >= max_j) or (k >= max_k)): 
        return

    t_min=tmp_min[0,1,2]
    t_max=tmp_max[0,1,2]

    #TODO:fare variante per ogni canale, mannacia
    results[i,j,k] = round(156* (values[i,j,k] -48 - t_min)/(t_max-2*t_min))#TODO:parametrizare il 48 e 56 dai dati del tflite


#kernel con multi filtro



@cuda.jit(cache=True)    
def convolve_tflite_mf(result,a,b,params,k_mul,k_filtri,k_bias,M,activation,out_mul,out_offest,stride,idx_multiply,K_p):

    x, y,p = cuda.grid(3) 
    
    p=K_p*p
    max_x=result.shape[0]
    max_y=result.shape[1]
    n_channels = params[0]
    if ((x >= max_x)  or (y >= max_y) or (p >=n_channels)): 
        return  

   
    offset1 = params[1]
    offset2 = params[2]
    xi = stride[0] * x
    yi = stride[1] * y

    range_i = range(xi,xi+b.shape[1])
    range_j = range(yi,yi+b.shape[2])
    range_p = range(0,K_p)
    
    s = cuda.local.array(shape=(16), dtype=int32)

    for pp in range_p:
        s[pp]=0
        #calcolo cella convoluzione
        #todo: aggiungere lo step
    for i in range_i:
        for j in range_j:
            for k in range(b.shape[3]):
                #s+=1
                A = int8(a[i,j,k]+offset1)
                for pp in range_p:
                    
                    s[pp] += M[int8(idx_multiply[pp+p]),A,int8(b[pp+p,i-xi,j-yi,k]+offset2)]

    for pp in range_p:
        ris = s[pp]*k_mul[pp+p] - k_filtri[pp+p] + k_bias[pp+p]

        if activation and ris<0:
            ris =0
        result[x,y,pp+p] = round(ris*out_mul + out_offest)

@cuda.jit(cache=True)  
def convolve_tflite_padding_mf(result, a,b,params,k_mul,k_filtri,k_bias,M,activation,out_mul,out_offest,value_null,stride,idx_multiply,K_p):

    x, y,p = cuda.grid(3)    
    max_x=result.shape[0]
    max_y=result.shape[1]
    n_channels = params[0]
    p=K_p*p
    if ((x >= max_x)  or (y >= max_y) or (p >=n_channels)): 
        return
    
  
    offset1 = params[1]
    offset2 = params[2]
    xi = stride[0] * x
    yi = stride[1] * y
    
    delta_i = int((b.shape[1]-1)/2)
    delta_j = int((b.shape[2]-1)/2)
    

    if(stride[0] > 1):
        offset_i=xi
        delta_i=2*delta_i
    else:
        offset_i=xi-delta_i
        

    if(stride[1] > 1):
        offset_j=yi
        delta_j=2*delta_j
    else:
        offset_j=yi-delta_j

    range_i = range(offset_i,xi+delta_i+1)
    range_j = range(offset_j,yi+delta_j+1)
    range_p = range(0,K_p)
    
        #calcolo cella convoluzion
        #todo: aggiungere lo step
    s = cuda.local.array(shape=(16), dtype=int32)

    for pp in range_p:
        s[pp]=0

    for i in range_i:
        for j in range_j:
            for k in range(b.shape[3]):
                #s+=1
                if i >= 0 and  i<a.shape[0] and j >= 0 and j < a.shape[1]:
                    A=int8(a[i,j,k]+offset1)
                    for pp in range_p:
                        s[pp] += M[int8(idx_multiply[pp+p]),A,int8(b[pp+p,i-offset_i,j-offset_j,k]+offset2)]
                else:
                    for pp in range_p:
                        A=int8(value_null+offset1)
                        s[pp] += M[int8(idx_multiply[pp+p]),A,int8(b[pp+p,i-offset_i,j-offset_j,k]+offset2)]

    for pp in range_p:
        ris = s[pp]*k_mul[pp+p] - k_filtri[pp+p] + k_bias[pp+p]

        if activation and ris<0:
            ris =0

        result[x,y,pp+p] =round(ris*out_mul + out_offest)


@cuda.jit(cache=True)  
def convolve_tflite_fake_mf(result, a,b,params,k_mul,k_filtri,k_bias,M,activation,out_mul,out_offest,stride,idx_multiply,K_p):

    #impone che la dimensione del kernel sia 1x1
    x, y,p = cuda.grid(3)    
    max_x=result.shape[0]
    max_y=result.shape[1]

    p=K_p*p
   
    
    n_channels = params[0]
    offset1 = params[1]
    offset2 = params[2]
    xi = stride[0] * x
    yi = stride[1] * y
    
    if ((x >= max_x)  or (y >= max_y) or (p >=n_channels)): 
        return

    range_p = range(0,K_p)
    
        #calcolo cella convoluzione
        #todo: aggiungere lo step
    s = cuda.local.array(shape=(16), dtype=int32)

    for pp in range_p:
        s[pp]=0


    for k in range(b.shape[3]):

        A=int8(a[xi,yi,k]+offset1)
        for pp in range_p:
            s[pp] += M[int8(idx_multiply[pp+p]),A,int8(b[pp+p,0,0,k]+offset2)]


    for pp in range_p:
        ris = s[pp]*k_mul[pp+p] - k_filtri[pp+p] + k_bias[pp+p]

        if activation and ris<0:
            ris =0

        result[x,y,pp+p] =round(ris*out_mul + out_offest)
        
        
#Convoluzione per fibre

@cuda.jit(cache=True)    
def convolve_tflite_mfiber(result,a,b,params,k_mul,k_filtri,k_bias,M,activation,out_mul,out_offest,stride,idx_multiply,K_p):
    
    x, y, p = cuda.grid(3) 
    max_x=result.shape[0]
    max_y=result.shape[1]
    
    p=K_p*p

    n_channels = params[0]
    
    if ((x >= max_x)  or (y >= max_y) or (p >=n_channels)): 
        return 
    
    offset1 = params[1]
    offset2 = params[2]
    xi = stride[0] * x
    yi = stride[1] * y

    range_i = range(xi,xi+b.shape[1])
    range_j = range(yi,yi+b.shape[2])
    range_p = range(0,K_p)
    
    s = cuda.local.array(shape=(16), dtype=int32)

    s=0
    #calcolo cella convoluzione
    #todo: aggiungere lo step
    for i in range_i:
        for j in range_j:
            for k in range(b.shape[3]):
                A = int8(a[i,j,k]+offset1)
                s += M[np.int8(idx_multiply[x][y]), A, np.int8(b[p, i - xi, j - yi, k] + offset2)]

    
        ris = s*k_mul[p] - k_filtri[p] + k_bias[p]

        if activation and ris<0:
            ris =0
        result[x,y,p] = round(ris*out_mul + out_offest)
        

# def convolve_tflite_mfiber_cpu(result, a, b, params, k_mul, k_filtri, k_bias, M, activation, out_mul, out_offest, stride, idx_multiply, K_p):
#     print("matrice mux:", idx_multiply.copy_to_host())
    
#     max_x = result.shape[0]
#     max_y = result.shape[1]

#     print(params.copy_to_host())
#     n_channels = params[0]

#     offset1 = params[1]
#     offset2 = params[2]

#     for x in tqdm(range(max_x)):
#         for y in range(max_y):
#             for p in range(n_channels):
#                 xi = stride[0] * x
#                 yi = stride[1] * y
                
#                 #print("xi:", xi,"yi:", yi)
#                 s = 0
                
#                 for i in range(xi,xi+b.shape[1]):
#                     for j in range(yi,yi+b.shape[2]):
#                         for k in range(b.shape[3]):
#                             A = np.int8(a[i, j, k] + offset1)
#                             s += M[np.int8(idx_multiply[x][y]), A, np.int8(b[p, i - xi, j - yi, k] + offset2)]
                       
#                 ris = s * k_mul[p] - k_filtri[p] + k_bias[p]

#                 if activation and ris < 0:
#                     ris = 0
#                 result[x, y, p] = round(ris * out_mul + out_offest)


#ANCORA NON TESTATO
@cuda.jit(cache=True)  
def convolve_tflite_fake_mfiber(result, a,b,params,k_mul,k_filtri,k_bias,M,activation,out_mul,out_offest,stride,idx_multiply,K_p):

    #impone che la dimensione del kernel sia 1x1
    x, y,p = cuda.grid(3)    
    max_x=result.shape[0]
    max_y=result.shape[1]

    p=K_p*p

    
    n_channels = params[0]
    offset1 = params[1]
    offset2 = params[2]
    xi = stride[0] * x
    yi = stride[1] * y
    
    if ((x >= max_x)  or (y >= max_y) or (p >=n_channels)): 
        return
    #range_p = range(0,K_p)
    
        #calcolo cella convoluzione
        #todo: aggiungere lo step
    s = cuda.local.array(shape=(1), dtype=int32)

    #s=0 

    for k in range(b.shape[3]):

        A=int8(a[xi,yi,k]+offset1)
        s[0] += M[int8(idx_multiply[x][y]),A,int8(b[p,0,0,k]+offset2)]


    ris = s[0]*k_mul[p] - k_filtri[p] + k_bias[p]

    if activation and ris<0:
        ris =0

    result[x,y,p] =round(ris*out_mul + out_offest)        
    
    
@cuda.jit(cache=True)  
def convolve_tflite_padding_mfiber(result, a,b,params,k_mul,k_filtri,k_bias,M,activation,out_mul,out_offest,value_null,stride,idx_multiply,K_p):

    x, y,p = cuda.grid(3)    
    max_x=result.shape[0]
    max_y=result.shape[1]
    n_channels = params[0]
    p=K_p*p
    if ((x >= max_x)  or (y >= max_y) or (p >=n_channels)): 
        return
    
  
    offset1 = params[1]
    offset2 = params[2]
    xi = stride[0] * x
    yi = stride[1] * y
    
    delta_i = int((b.shape[1]-1)/2)
    delta_j = int((b.shape[2]-1)/2)
    

    if(stride[0] > 1):
        offset_i=xi
        delta_i=2*delta_i
    else:
        offset_i=xi-delta_i
        

    if(stride[1] > 1):
        offset_j=yi
        delta_j=2*delta_j
    else:
        offset_j=yi-delta_j

    range_i = range(offset_i,xi+delta_i+1)
    range_j = range(offset_j,yi+delta_j+1)
    range_p = range(0,K_p)
    
        #calcolo cella convoluzion
        #todo: aggiungere lo step
    s = cuda.local.array(shape=(16), dtype=int32)

    #for pp in range_p:
    s=0

    for i in range_i:
        for j in range_j:
            for k in range(b.shape[3]):
                #s+=1
                if i >= 0 and  i<a.shape[0] and j >= 0 and j < a.shape[1]:
                    A=int8(a[i,j,k]+offset1)
                    s += M[int8(idx_multiply[x][y]),A,int8(b[p,i-offset_i,j-offset_j,k]+offset2)]
                else:
                    A=int8(value_null+offset1)
                    s += M[int8(idx_multiply[x][y]),A,int8(b[p,i-offset_i,j-offset_j,k]+offset2)]

    ris = s*k_mul[p] - k_filtri[p] + k_bias[p]

    if activation and ris<0:
        ris =0

    result[x,y,p] =round(ris*out_mul + out_offest)      
                


import time
from numba import cuda
import numpy as np
from inspectnn.Model.BaseModel import BaseModel
class BaseModel_tflite(BaseModel):
    def __init__(self, multiplier,quant_nbits = 8,input_shape=[28, 28, 1],save_csv_path=''):
        
        super().__init__(multiplier,quant_nbits,input_shape,save_csv_path)

        self.elapsed_time = 0
        #self.interpreter = []
        self.total_elapsed_time=0
        self.baseline_accuracy = None
        
        self.net = None
         
    def evaluate2(self,images,labels,log=False):
        num_of_samples=len(images)
        if log : 
            print(f"Testing on {len(images)} images.")
        st = time.time()
        Accuracy = self.net.evaluate_accuracy(labels, images)
        if log :
            print(f"Accuracy: {Accuracy}")
        et = time.time()
        self.elapsed_time = et - st
        self.total_elapsed_time+=self.elapsed_time
        if log :
            print('MIMT execution time:', self.elapsed_time, 'seconds')
            print('FPS:', num_of_samples/self.elapsed_time) 
            self.net.print_time_statics()
        return Accuracy
    

    def evaluate_quantized_model(self, x_test_set, y_test_set):
        input_index = self.interpreter.get_input_details()[0]["index"]
        output_index = self.interpreter.get_output_details()[0]["index"]
        prediction_digits = []
        labels = []
        for i, test_image in enumerate(x_test_set):
            test_image = np.expand_dims(test_image, axis=0).astype(np.float32)
            self.interpreter.set_tensor(input_index, test_image)
            # Run inference.
            self.interpreter.invoke()

            output = self.interpreter.tensor(output_index)
            digit = np.argmax(output()[0])
            prediction_digits.append(np.array(digit))
            labels.append(y_test_set[i])

        prediction_digits = np.array(prediction_digits)
        accuracy = (prediction_digits == np.array(labels)).mean()
        return accuracy
    
    def evaluate_tflite(self, x_test_set, y_test_set,print_debugs=None,print_acc=True):

        input_index = self.interpreter.get_input_details()[0]["index"]
        output_index = self.interpreter.get_output_details()[0]["index"]
        prediction_digits = []
        labels = []
        #print(self.interpreter.get_input_details()[0])
        start_time = time.time()
        for i, test_image in enumerate(x_test_set):
            
            test_image = np.expand_dims(test_image, axis=0).astype(np.float32)#fp32
            #print("test image shape: " ,test_image.shape)
            self.interpreter.set_tensor(input_index, test_image)

            # Run inference.
            self.interpreter.invoke()
            
            if print_debugs is not None:
                #valori resnet8
                quant_layer=6
                conv_lay = 9
                conv1_lay = 12
                for print_debug in print_debugs:
                    print('### tflite interpetrer: -------#####>>>>>> L(',print_debug,")")
                    print("NAME: ",self.interpreter.get_tensor_details()[print_debug]['name'].split('/')[1])
                    print(self.interpreter.get_tensor(print_debug).shape)
                    #print("max:",np.amax(self.interpreter.get_tensor(print_debug)))
                    print(self.interpreter.get_tensor(print_debug)[0][0][0])

                


            output = self.interpreter.tensor(output_index)
            digit = np.argmax(output())
            prediction_digits.append(np.array(digit))
            labels.append(int(y_test_set[i]))
            #print(digit,labels[-1])
        delta_time = time.time() - start_time
        
        #print(prediction_digits)
        prediction_digits = np.array(prediction_digits)     
        accuracy = (prediction_digits == np.array(labels)).mean()
        if print_acc:
            #print(f"Accuracy Tflite: {accuracy * 100:.2f}%, {len(prediction_digits)} images in {delta_time}s - {len(prediction_digits)/delta_time} FPS ")
            print(f"Accuracy Tflite: {accuracy * 100:.2f}%, {len(prediction_digits)} images in {round(delta_time,2)}s - {round(len(prediction_digits)/delta_time,2)} FPS ")
       
        return accuracy
    
   
        

    def trova_dif(self, x_test_set_tf, y_test_set_tf):
        counter = 0
        print("Trova DIFF ....")
        self.net.update_multipler([self.exact_multiplier])
        for i in range(0,len(x_test_set_tf)):
            
            predict_class = self.net.predict(self.images[i])
            nostro_tool=np.argmax(predict_class) == self.labels[i]#.append()

            tflite = self.evaluate_tflite(x_test_set_tf[i:i+1],y_test_set_tf[i:i+1],print_acc=False)
            
            if(nostro_tool!=tflite):
                print(counter,i,nostro_tool,sep='\t')
                counter+=1
        
        return counter/len(x_test_set_tf)
    
    
 
        



                

    

    
    
    
        


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import numpy as np
from numba import cuda, jit
from inspectnn.BaseLayer import BaseLayer
from inspectnn.Reshape.Reshape_kernel_tflite import reshape_3d

class ReshapeLayer(BaseLayer):
    def __init__(self, name = "Reshape",dim_size = [1,1]):
        #TODO: Fare meglioo l' init
        super().__init__(self, name = name)
        self.dim_reshape = dim_size
        
    def __deepcopy__(self, memo = None):
        return ReshapeLayer(name = self.name,dim_size=self.dim_reshape)
    
    def load_weights(self, **kwargs):
        
        self.input_shape, self.enable_gpu = kwargs["input_shape"],kwargs["enable_gpu"] 
        self.output_shape = kwargs["output_shape"]      
        
        if self.enable_gpu:
            self.use_gpu=True

            
            self.use_3d = True
            self.blockdim = (2,2,2)
            self.griddim = (self.input_shape[0] // self.blockdim[0] + 1, self.input_shape[1] // self.blockdim[1] + 1,self.input_shape[2] // self.blockdim[2] + 1)#,n_channels)
        

            self.results = cuda.device_array(self.output_shape, dtype=np.int8)#int16 o 8
            
            self.dim_reshape_gpu = cuda.to_device(np.ascontiguousarray(self.dim_reshape))
            self.gpu_input_memory = kwargs["gpu_input_memory"]
            self.output_mul= 1/kwargs["quant_output_k"]
            self.output_offset= kwargs["quant_output_offset"]

        return self.output_shape
    
    #@jit
    def forward_pass(self):
        if self.gpu_input_memory is None:
            print("flat no buoeno")
            
        else:
            # TODO reshape su GPU
            cuda.synchronize()
            reshape_3d[self.griddim, self.blockdim](self.results, self.gpu_input_memory,self.dim_reshape_gpu)
            
            #print(self.name,"out",self.results.copy_to_host()[4][4])
            
            
 

        


from sklearn.model_selection import train_test_split


import os,pandas
import numpy as np
from tensorflow.keras.preprocessing.image import ImageDataGenerator

my_PATH = os.path.dirname(__file__)+ '/CaniGatti'

def GattidmvsCani(num_of_samples,sequenzial=False,PATH=my_PATH):
    

    animals = os.listdir(PATH)
    
    images = [os.path.join(PATH, animal) for animal in animals]
    
    label = []
    categories = []
    for animal in images:
        if "Cat" in animal:
            categories.append('0')
        elif "Dog" in animal:
            categories.append('1')
            
        else:
            print(animal + " non identificato!")
            
    animals = pandas.DataFrame({'Image_name': np.array([f'{f}' for f in images]), 'Category': categories})
    
    if num_of_samples > 0:
        batch_size = num_of_samples
    else:
        batch_size = 1-num_of_samples
        
        
    test_datagen = ImageDataGenerator(rescale=1./255)
    test_generator = test_datagen.flow_from_dataframe(animals, x_col='Image_name', y_col='Category', class_mode = 'binary', target_size = (224, 224), batch_size = batch_size, validate_filenames = False,shuffle=not sequenzial)
    y_test = animals['Category']
    len_test_dataset = animals.shape[0]
    
  
    # for i in range(len(test_generator)):
    #    batch_images, batch_labels = test_generator[i]
    #    for j in range(batch_images.shape[0]):
    #        image = batch_images[j]
    #        label = batch_labels[j]

    if(num_of_samples > 0):

        img, lab = test_generator[0]
        #print(lab)
        # for i in range(1,num_of_samples):
        #     batch_images, batch_labels = test_generator[i]
        #     img.concatenate(batch_images)
        #     lab.concatenate(batch_labels)
        
        
        label = [int(j) for j in lab]
        return  None,label ,img, lab
    
    
    else:
        batch_images, batch_labels = test_generator[0]
        
        label = [int(batch_labels[-num_of_samples])]
        return  None,label,[batch_images[-num_of_samples]], [batch_labels[-num_of_samples]]


from sklearn.model_selection import train_test_split
import numpy as np, tensorflow as tf, random

def preprocess_image_input(input_images):
    input_images = input_images.astype('float32')
    output_ims = tf.keras.applications.resnet50.preprocess_input(input_images)
    return output_ims

    
def Dataset_cifar100(num_of_samples,sequenzial=False,PATH=None):
    
    (x_train, y_train), (x_test, y_test) = tf.keras.datasets.cifar100.load_data()
    x_train, x_val, y_train, y_val = train_test_split(x_train, y_train, test_size=0.1, random_state=1)

    # Preprocess the data
    x_train_a = x_train.astype('float32')
    x_test_a = x_test.astype('float32')
    x_val_a = x_val.astype('float32')

    x_train_a /= 255
    x_test_a /= 255
    x_val_a /= 255

    x_test = np.uint8(x_test)
    if(num_of_samples > 0):
        if sequenzial :
            dataset = range( num_of_samples)
        else:
            dataset = random.sample(range(len(x_test_a)), num_of_samples)
    else:
        dataset = [-num_of_samples]

    if(num_of_samples == 1):
        print("Dataset:",dataset)
    #print(x_test.shape)
    
    #dataset= [15]
    c10_images = [x_test_a[i] for i in dataset ]
    c10_labels = [y_test[i][0] for i in dataset]

    c10_images_apx = [x_test[i] for i in dataset ]
    c10_labels_apx = c10_labels

    return c10_images_apx,c10_labels_apx,c10_images,c10_labels


  
def Dataset_cifar100_resnet50(num_of_samples,sequenzial=False,PATH=None):
    
    (x_train, y_train), (x_test, y_test) = tf.keras.datasets.cifar100.load_data()
    x_train, x_val, y_train, y_val = train_test_split(x_train, y_train, test_size=0.1, random_state=1)


    x_train_a = preprocess_image_input(x_train)
    x_test_a = preprocess_image_input(x_test)
    x_val_a = preprocess_image_input(x_val)

    x_test = np.uint8(x_test)
    if(num_of_samples > 0):
        if sequenzial :
            dataset = range( num_of_samples)
        else:
            dataset = random.sample(range(len(x_test_a)), num_of_samples)
    else:
        dataset = [-num_of_samples]

    if(num_of_samples == 1):
        print("Dataset:",dataset)
    #print(x_test.shape)
    
    #dataset= [15]
    c10_images = [x_test_a[i] for i in dataset ]
    c10_labels = [y_test[i][0] for i in dataset]

    c10_images_apx = [x_test[i] for i in dataset ]
    c10_labels_apx = c10_labels

    return c10_images,c10_labels_apx,c10_images,c10_labels

def load_percentage_images_cf100(v,num_of_samples=10000,PATH=None):
    (x_train, y_train), (x_test, y_test) = tf.keras.datasets.cifar100.load_data()
    x_train, x_val, y_train, y_val = train_test_split(x_train, y_train, test_size=0.1, random_state=1)

    # Preprocess the data
    x_train_a = x_train.astype('float32')
    x_test_a = x_test.astype('float32')
    x_val_a = x_val.astype('float32')

    x_train_a /= 255
    x_test_a /= 255
    x_val_a /= 255

    x_test = x_test.astype('int8')
    dataset = range( num_of_samples)
    
    c10_images = [x_test_a[i] for i in dataset ]
    c10_labels = [y_test[i][0] for i in dataset]
    c10_images_apx = [x_test[i] for i in dataset ]

    result_images_apx=[]
    result_images = []
    result_labels = []
    
    for id,l in enumerate(c10_labels):
        if(v[l]>0):
            v[l]-=1
            result_images_apx.append(c10_images_apx[id])
            result_images.append(c10_images[id])
            result_labels.append(l)

    return result_images_apx,result_labels,c10_images,result_labels



"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import numpy as np
import time

from inspectnn.Dense.Dense_cuda import Dense_cuda

from numba import cuda
from inspectnn.BaseLayer import BaseLayer
from inspectnn.Dense.Dense_kernel_tflite import matmul_tflite,matmul_tflite_occorenza,matmul_tflite_mf,matmul_tflite_fast_pippo

from inspectnn.Dense.DenseLayer import DenseLayer



class DenseLayer_TFLITE(DenseLayer):
    total_type_layer_time = 0
    def __init__(self, activation = "relu", quant_nbits = 8, multiplier = BaseLayer.default_multiplier, name = "layer",offset=[128,128],print_output=None):
        super().__init__( activation, quant_nbits, multiplier, name,print_output)
        self.offset=offset
        self.idx_multiply = None

    def __deepcopy__(self, memo = None):
        return DenseLayer_TFLITE(activation = self.activation, quant_nbits = self.quant_nbits, multiplier = self.multiplier, name = self.name)

    #@jit
    def forward_pass(self):
        # TODO Aggiungi supporto per altre attivazioni
    
        #print(self.idx_multiply)

        #sprint(self.name,"input\n", self.pre_layer.results.copy_to_host()[0][:16])

        if(self.idx_multiply is None):
            self.test_cuda.cuda_set( self,self.pre_layer.results)
            
            #self.cuda_set( A_global_mem )
            #cuda.synchronize()
            #matmul_tflite
            #fast_matmul_tflite
            #print([self.griddim, self.blockdim])
            #print('ris_mul:',self.results.copy_to_host())
            self.test_cuda.MatrixRowsGPU()
            #print('ris_mul:',self.results.copy_to_host())
            
            #cuda.synchronize()
            #matmul_tflite[self.griddim, self.blockdim](self.results, A_global_mem,self.weights,self.k_mul,self.k_filtri,self.k_bias,self.M,128,128,self.activation,self.output_mul,self.output_offset,self.update_bias)
            #print('ris_mul:',self.results.copy_to_host())
        else:
            #print(self.name,self.idx_multiply)
            #idxm = cuda.to_device(np.ascontiguousarray(self.idx_multiply))
            cuda.synchronize()
            matmul_tflite_mf[self.griddim, self.blockdim](self.results, self.pre_layer.results,self.weights,self.k_mul,self.k_filtri,self.k_bias,self.all_M,128,128,self.activation,self.output_mul,self.output_offset,idxm)
    

        #print('ris_mul:',self.results.copy_to_host())
        #TODO: ottimizare questo schifo di if (è tempo di portare la softmax sulla gpu?- ancora no)
    
        if self.gpu_output_memory == False:
            cuda.synchronize()
            self.outputv[:,:] = self.results.copy_to_host()
            
    def load_input_layer(self,layers,tensors_out):
        
        super().load_input_layer(layers,tensors_out)
        self.test_cuda.cuda_set(self,self.pre_layer.results )  

    def load_weights(self, **kwargs):
        self.test_cuda = Dense_cuda()
        self.kernel_shape = np.shape(kwargs["weights"])
        self.enable_gpu = kwargs["enable_gpu"]
        self.input_shape = kwargs["input_shape"]
        self.weights_cpu=np.array(kwargs["weights"],dtype=np.int16)
        self.weights_shape = np.shape(kwargs["weights"])
        #print("weights_shape",self.weights_shape)
        #self.weights=kwargs["weights"]
        self.output_shape = kwargs["output_shape"]  
        self.outputv = np.zeros(self.output_shape)
        self.blockdim = (1,8)
        self.n_filter = self.output_shape[1]
        self.griddim = (self.output_shape[0] // self.blockdim[0] + 1, self.output_shape[1] // self.blockdim[1] + 1)#,n_channels)
        if self.enable_gpu:
            self.use_gpu=True
            self.M = self.multiplier
            #todo accorciare
            self.weights = cuda.to_device(np.array(kwargs["weights"],dtype=np.int8))
            #self.occorenza = cuda.to_device(np.zeros((self.n_filter,256,256),dtype=np.int32))
            self.results = cuda.device_array(self.output_shape, dtype=np.int8)         
            self.gpu_input_memory = kwargs["gpu_input_memory"]

            if ( kwargs["quant_data"] is not None):
                self.quantization=True
                self.quanto_bias = cuda.to_device(np.ascontiguousarray(kwargs["quant_bias"]))
                self.quanto_B = cuda.to_device(np.ascontiguousarray(kwargs["quant_data"]))
                
                self.update_bias = cuda.to_device(np.ascontiguousarray(np.zeros(self.n_filter)))
                self.k_mul = cuda.to_device(np.ascontiguousarray(kwargs["quant_data"]*kwargs["quant_input_k"],dtype=np.single))#TODO:parametrizare il 255
                self.k_filtri = cuda.to_device(np.ascontiguousarray(np.sum(kwargs["weights"],(1))*kwargs["quant_data"]*kwargs["quant_input_offset"]*kwargs["quant_input_k"],dtype=np.single))#TODO:parametrizare il diviso 2 con alfa input+offest input
                self.k_bias = cuda.to_device(np.ascontiguousarray(kwargs["biases"]*kwargs["quant_bias"],dtype=np.single))

                self.output_mul= 1/kwargs["quant_output_k"]
                self.output_offset= kwargs["quant_output_offset"]

        self.activation = (self.activation == "relu" or self.activation == "Relu")
        if self.gpu_input_memory is None:
            self.outputv = np.zeros(self.output_shape)   
        return self.output_shape
   
    
    def research_hyperparameters(self):
        
        best_blockdim = self.blockdim
        best_griddim = self.griddim 
        best_time = 100000000
        for i in range(0,16):
            for j in range(0,16):
            
                self.blockdim = (i+1,j+1)
                self.griddim = (self.output_shape[0] // self.blockdim[0] + 1, self.output_shape[1] // self.blockdim[1] + 1)#,n_channels)
                init_time = time.time()
                
                for i in range(2):
                    self.forward_pass()
                    
                tot_time = time.time() - init_time
                #print("blockdim:",self.blockdim,"griddim:",self.griddim,"time:",tot_time)
                
                if tot_time < best_time:
                    best_blockdim = self.blockdim
                    best_griddim = self.griddim
                    best_time = tot_time
        
        print("best_blockdim:",best_blockdim,"best_griddim:",best_griddim)
        
        self.blockdim = best_blockdim
        self.griddim = best_griddim
        


"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import numpy as np, json, copy, time,os,tempfile,mmap
from numba import cuda, jit

from distutils.dir_util import mkpath
from inspectnn.Conv.ConvLayer_TFLITE import ConvLayer
from inspectnn.Conv.ConvLayer_TFLITE import ConvLayer_TFLITE
from inspectnn.Dense.DenseLayer_TFLITE import DenseLayer
from inspectnn.Dense.DenseLayer_TFLITE import DenseLayer_TFLITE
from inspectnn.Pooling.PoolingLayer import PoolingLayer
from inspectnn.Pooling.PoolingLayer_TFLITE import PoolingLayer_TFLITE
from inspectnn.Flatten.FlattenLayer import FlattenLayer
from inspectnn.Quantizate.QuantizateLayer import QuantizateLayer
from inspectnn.Add.AddLayer import AddLayer
from inspectnn.PAD.PadLayer import PadLayer
#from .OutputLayer import OutputLayer
from inspectnn.Utils.utils import grouped, flatten


class NetworkModel:
    def __init__(self, learned_parameters, layers,input_shape,print_input=False,quant_data=None,quant_data_scale=None):
        self.learned_parameters = learned_parameters
        self.layers = layers
        self.outputs = []
        self.input_shape=input_shape
        self.print_input = print_input
        self.total_time = 0
        self.n_frame_inference = 0
        self.elapsed_time = np.zeros(len(layers))
        self.quant_data=quant_data
        self.N_layer_with_multiply = 0
       
        gpu_input_memory=None
        
        lp = 0
        last_layer=None
        disponibilita_gpu=True#tf.test.is_gpu_available()

        quantization_bias=None
        quantization_pesi=None
        quant_input_k=None
        quant_input_offset=None
        quant_output_k = None
        quant_output_offset=None
        self.init_quantizate = True
        
        #pre_quant_data_scale = DATA_layer_quant(1/255,-128)
        for id_layer in range(len(self.layers)):
            
            if isinstance(self.layers[id_layer], (ConvLayer, DenseLayer)):
                if(self.quant_data != None):

                    quantization_pesi = quant_data[lp].scales
                    quantization_bias = quant_data[lp+1].scales
                    quant_input_k = pre_quant_data_scale.scales
                    quant_input_offset = pre_quant_data_scale.offset
                    quant_output_k = quant_data_scale[lp//2].scales
                    quant_output_offset = quant_data_scale[lp//2].offset
                   
                    pre_quant_data_scale = quant_data_scale[lp//2]

               
                output_shape =self.layers[id_layer].load_weights(input_shape = input_shape, weights = self.learned_parameters[lp], biases = self.learned_parameters[lp + 1],
                                                                        enable_gpu=disponibilita_gpu,gpu_input_memory=gpu_input_memory,pre_layer=last_layer,
                                                                        quant_data=quantization_pesi,quant_bias=quantization_bias,quant_input_k=quant_input_k,
                                                                        quant_input_offset=quant_input_offset,quant_output_k=quant_output_k,quant_output_offset=quant_output_offset)
                lp += 2
            
                
            elif isinstance(self.layers[id_layer], (PoolingLayer, FlattenLayer, QuantizateLayer)):
                output_shape =self.layers[id_layer].load_weights(input_shape = input_shape,enable_gpu=disponibilita_gpu,gpu_input_memory=gpu_input_memory,pre_layer=last_layer)
                
            else:
                print(f"Layer {self.layers[id_layer].name} not supported yet")
            
            gpu_input_memory=self.layers[id_layer].results
            if self.layers[id_layer].use_gpu  and   last_layer != None and last_layer.use_gpu:
                last_layer.gpu_output_memory = True
            else:
                print(f"Layer {self.layers[id_layer].name} no gpu")

            input_shape = output_shape
            self.layers[id_layer].pre_layer = last_layer
            last_layer = self.layers[id_layer]
        

    def __deepcopy__(self, memo = None):
        return NetworkModel(self.learned_parameters, copy.deepcopy(self.layers),self.input_shape)

    def export_learned_parameters(self):
        data = []
        for i, p in enumerate(grouped(self.learned_parameters, 2)):
            tmp = list(flatten(p[0].astype(int).tolist()))
            with open(f"layer_{i}.json", 'w') as outfile:
                outfile.write(json.dumps(tmp))
            data.append(tmp)
        data = list(flatten(data))
        with open("network.json", 'w') as outfile:
            outfile.write(json.dumps(data))
                
    def update_multipler(self,multipliers,layer_apx=None):
        idx=0

        if layer_apx is None:
            n_layer_apx = len(self.layers)
        else:
            n_layer_apx = layer_apx
        
        #modifica epr tutti i layer
        for id_layer in range(n_layer_apx):
            if isinstance(self.layers[id_layer], (ConvLayer, DenseLayer)):
                
                if(len(multipliers)==1):
                    self.layers[id_layer].M = multipliers[0].multiplier #[idx]
                    #stampa l' indirizzo in esadecimale
                    
                    #3if isinstance(self.layers[id_layer], (ConvLayer)):#TODO: add DenseLayer
                    self.layers[id_layer].test_cuda.setPointerMul( self.layers[id_layer].M.device_ctypes_pointer.value)
                        
                    self.layers[id_layer].multiplier = multipliers[0]
                else:
                    
                    self.layers[id_layer].multiplier = multipliers[idx]
                    
                    #print(f"Moltiplicatori per il layer {self.layers[id_layer]}: ", self.layers[id_layer].multiplier)
                    #TODO: aggiungere update_bias_for_mul per ogni layer conv e dense
                    #QUesto valore è un vector di bias per per ogni moltiplier
                    #self.layers[id_layer].update_bias = self.layers[id_layer].update_bias_for_mul[idx]
                    
                    self.layers[id_layer].M = multipliers[idx].multiplier
                    self.layers[id_layer].test_cuda.setPointerMul( self.layers[id_layer].M.device_ctypes_pointer.value)
                    idx+=1
       
    def get_number_of_all_filters(self, recalculated=False):
        if recalculated or self.n_filter is None:
            self.n_filter=0
            for id_layer in range(len(self.layers)):
                if isinstance(self.layers[id_layer], (ConvLayer, DenseLayer)):
                    print(f"[{id_layer}]Layer {self.layers[id_layer].name} n_filter {self.layers[id_layer].n_filter}")
                    self.n_filter+=self.layers[id_layer].n_filter
        
        return self.n_filter
    
    def update_multipler_filters(self,multipliers, dense_multiplier):#TODO: add for cluster
        
        
        
        cont_conv=0
        dense_cont = 0
        for id_layer in range(len(self.layers)):
            if isinstance(self.layers[id_layer], (ConvLayer)):
                
                n_filter = self.layers[id_layer].n_filter
                #print("nfilter", n_filter)
                self.layers[id_layer].idx_multiply = multipliers[cont_conv:(cont_conv+n_filter)]
                #print(f"Moltiplicatori per il layer {self.layers[id_layer]}: ", self.layers[id_layer].idx_multiply)
                #print(multipliers[idx:(idx+n_filter)])
                cont_conv += n_filter
                
            elif isinstance(self.layers[id_layer], (DenseLayer)):
                
                n_filter = self.layers[id_layer].n_filter
                #print("Numero di filtri:", n_filter)
                self.layers[id_layer].idx_multiply = dense_multiplier[dense_cont:(dense_cont+n_filter)]
                #print(f"Moltiplicatori per il layer {self.layers[id_layer]}: ", self.layers[id_layer].idx_multiply)
                
                dense_cont += n_filter
                
                
    
    
    #Call this function to approximate all Conv Layer by fibres. You need to specify all multipliers for all layer 
    def update_multiplier_for_fibers(self,fibers_multiplier_id, dense_multiplier):#TODO: add for cluster
        
        # print("Multiplier Conv")
        # for elem in fibers_multiplier_id:
        #     print(elem)
            
        # print("Multiplier Dense")
        # for elem in dense_multiplier:
        #     print(dense_multiplier)

        
        
        
        #!NOTA: fibers_multiplier_id è la matrice di id di moltiplicatori per fibra per i layer a convoluzione
        #!      dense_multiplier è l'oggetto moltiplicatore per i layer dense.
        
        conv_counter = 0
        dense_counter = 0
        for id_layer in range(len(self.layers)):
            if isinstance(self.layers[id_layer], (ConvLayer)):
                
                self.layers[id_layer].mul_per_fibers = fibers_multiplier_id[conv_counter]
                
                #print(f"Moltiplicatori per il layer {self.layers[id_layer]}: ", self.layers[id_layer].mul_per_fibers)
                conv_counter += 1
                
            elif isinstance(self.layers[id_layer], (DenseLayer)):
                
                
                
                self.layers[id_layer].multiplier = dense_multiplier[dense_counter]
            #     #TODO: aggiungere update_bias_for_mul per ogni layer conv e dense
            #     #QUesto valore è un vector di bias per per ogni moltiplier
            #     #self.layers[id_layer].update_bias = self.layers[id_layer].update_bias_for_mul[idx]
                self.layers[id_layer].M = dense_multiplier[dense_counter].multiplier
                
                #print(f"Moltiplicatori per il layer {self.layers[id_layer]}: ", dense_multiplier[dense_counter])
                dense_counter +=1
        
    
    #Call this function to approximate a specifi Conv layer by fibres. You need to specify all multipliers for the layer 
    def update_multiplier_fibers_by_layer():
        #if isinstance(self.layers[id_layer], (ConvLayer)): #!TODO: Add Dense Layer
        #    fibers = self.layers[id_layer].fibers
            
        return
    
    

    def update_multipler_filter(self,multipliers,id_layer,cluster = None):
        if cluster is None:
            if isinstance(self.layers[id_layer], (ConvLayer, DenseLayer)):
                
                n_filter = self.layers[id_layer].n_filter
                self.layers[id_layer].idx_multiply = multipliers
                #print(multipliers[idx:(idx+n_filter)])
            else:
                print(f"Layer{ self.layers[id_layer].name} not valid")
        else:
            if isinstance(self.layers[id_layer], (ConvLayer, DenseLayer)):
                n_filter = self.layers[id_layer].n_filter
                multipliers_for_layer = np.zeros(n_filter)
                for i in range (n_filter):
                    multipliers_for_layer[i] = multipliers[cluster[i]]
                
                self.layers[id_layer].idx_multiply = multipliers_for_layer
                #print(multipliers[idx:(idx+n_filter)])
            else:
                print(f"Layer{ self.layers[id_layer].name} not valid")

    def load_total_multiply_gpu(self,all_multipliers):
        
        for id_layer in range(len(self.layers)):
            if isinstance(self.layers[id_layer], (ConvLayer, DenseLayer)):
                
                self.layers[id_layer].all_M = all_multipliers
            
    def total_filter(self):
        total_filter=0
        for id_layer in range(len(self.layers)):
            if isinstance(self.layers[id_layer], (ConvLayer, DenseLayer)):
                total_filter+=self.layers[id_layer].n_filter
        
        print("Total filter:",total_filter)

        return total_filter
    
    def predict(self, test_image, enable_multiprocess = False):
        self.outputs = [test_image]
        self.n_frame_inference+=1
        print("Predict:", test_image.copy_to_host()[0][0])
        if self.print_input:
            with open('test_image.txt', 'w', encoding='UTF8', newline='') as f:
                print(self.outputs,file=f)
            
        for id_layer in range(len(self.layers)):
            
            st = time.time()
            if(self.print_input):
                self.layers[id_layer].update_save_path()
            if isinstance(self.layers[id_layer], (ConvLayer, DenseLayer,PoolingLayer, FlattenLayer,QuantizateLayer)):
                self.outputs =       [self.layers[id_layer].forward_pass(inputv = self.outputs[-1])]
            else:
                print(f"Layer {self.layers[id_layer].name} not supported yet")
            
            et = time.time()
            self.elapsed_time[id_layer] += et - st
            self.total_time += et - st

            
        
        return self.outputs[-1] / np.max(self.outputs[-1])
        
    def print_time_statics(self):
        print(f"Total Layers Time = {self.total_time}")
        print(f"Total FPS ={self.n_frame_inference/self.total_time}")
        idx=0
        for layer in self.layers:
            line_new = 'Layer {:>30}\tTime:{:>8}s ({:>5}% )'.format(layer.name,round(self.elapsed_time[idx], 2),round(self.elapsed_time[idx]/self.total_time*100,1))
            print(line_new,f"\tGPU:{layer.use_gpu}\tGPU_OUT:{layer.gpu_output_memory}")
            idx+=1
            
            
            
    def print_time_statics_v2(self,print_all_layer=True):
        self.fps = self.n_frame_inference/self.total_time
        print(f"Total Layers Time = {round(self.total_time,2)}s\t [{round(self.fps,1)} FPS]")
        n_total_multiply = 0
        for layer in self.layers:
            n_total_multiply =  n_total_multiply + layer.n_moltiplicazioni
        
        print(f"Total multiply = {n_total_multiply} --> {round(self.n_frame_inference*n_total_multiply/(1000000000*self.total_time),1)} GOPS")
        idx=0
        if(print_all_layer):
            for layer in self.layers:
                line_new = '[{:>7}] {:>30}\tTime:{:>8}s ({:>5}%)  Power {:>8} mW ({:>5}%) '.format(layer.layer_name,layer.name,round(self.elapsed_time[idx], 2),round(self.elapsed_time[idx]/self.total_time*100,1),
                                                                                                round(layer.power,1),round(100*layer.n_moltiplicazioni/n_total_multiply,2))
                print(line_new)
                idx+=1
    
    def print_type_time(self):
        type_l = [ConvLayer_TFLITE, DenseLayer_TFLITE,PoolingLayer_TFLITE, FlattenLayer,QuantizateLayer,AddLayer,PadLayer]
        total= 0
        for i in type_l:
            total+=i.total_type_layer_time
        
        print(f"Total Time: {round(total,2)}s")
        for i in type_l:
              print(f"{str(i.layer_name).ljust(15)} \t {round(i.total_type_layer_time,2)}s \t( {str(round(100*i.total_type_layer_time/total,2)).ljust(5)}% )")
        
        
            
    def mimt_do_inference(self, labels, images):
        assert isinstance(labels, (list, tuple))
        assert isinstance(images, (list, tuple))
        assert len(labels) == len(images)
        result=[]
        for l, i in zip(labels, images):
            result_predict = self.predict(i)[0]
            result.append(np.argmax(result_predict) == l)

        return result
        
    def evaluate_accuracy(self, labels, images):
        assert isinstance(labels, (list, tuple)), f"labels must be a list or tuple. labels is {type(labels)}"
        assert isinstance(images, (list, tuple)), f"images must be a list or tuple. labels is {type(images)}"
        assert len(labels) == len(images), f"images and labes must be the same size. len(labels): {len(labels)}, len(images): {len(images)}"

        result=0

        for i in range(len(images)):
            #result+=1
            #print("evaluate: ",images[i].copy_to_host())
            predict_class = self.predict(images[i])
            result+=np.argmax(predict_class) == labels[i]#.append()
        
        return 100*result / len(labels)
    
    def evaluate_accuracyv2(self, labels, images,threshold=0.2,K_ratio=10):
        assert isinstance(labels, (list, tuple))
        assert isinstance(images, (list, tuple))
        assert len(labels) == len(images)

        result=0#[]

        for i in range(len(images)//K_ratio):
            #result+=1
            result+=np.argmax(self.predict(images[i])) == labels[i]#.append()
        
        if result > threshold*len(images)//K_ratio :
            for i in range(len(images)//K_ratio,len(images)):
                #result+=1
                result+=np.argmax(self.predict(images[i])) == labels[i]#.append()

            return 100*result / len(labels)
        
        else:
            return 100*result / (len(labels)//K_ratio)
        

    def evaluate_confusion_matrix(self, images,labels,n_class=10):

        result_matrix=np.zeros((n_class,n_class),dtype=int)

        for i in range(len(images)):
            #result+=1
            predict_class = np.argmax(self.predict(images[i]))
            result_matrix[predict_class,labels[i]] +=1
            #result=np.argmax(predict_class) == labels[i]#.append()

        result_matrix_norm=result_matrix/np.sum(result_matrix,0)
        return result_matrix_norm
    

    def evaluate_delta_gradient(self, images1, images2):
        
        result=0#[]

        predict_class1 = self.predict(images1).copy()
        predict_class2 = self.predict(images2).copy()

        return predict_class1 - predict_class2
    

    def get_Numbers_layer_with_multiply(self):
        self.N_layer_with_multiply = 0
        for id_layer in range(len(self.layers)):
            if isinstance(self.layers[id_layer], (ConvLayer, DenseLayer)):
                self.N_layer_with_multiply+=1

        return self.N_layer_with_multiply
    
    def return_Area(self):
        total_area= 0
        for layer in self.layers:

            if isinstance(layer, (ConvLayer, DenseLayer)):
                total_area +=  layer.multiplier.Area
                
        return total_area
    
    # def return_Power(self):
    #     total_power = 0
    #     for layer in self.layers:
    #         if isinstance(layer, (ConvLayer, DenseLayer)):
    #             layer.power = 0
                
    #             layer.power = layer.multiplier.Power*layer.n_moltiplicazioni
    #             print("LayerPower: ", layer.power, "n_moltiplicazioni: ", layer.n_moltiplicazioni, "multiplier_power: ", layer.multiplier.Power )
                
    #             total_power +=  layer.power
                
    #     return total_power
    
    
    def return_Power(self):
        total_power = 0
        for layer in self.layers:
            if isinstance(layer, (ConvLayer, DenseLayer)):
                layer.power = layer.multiplier.Power*layer.n_moltiplicazioni
                total_power +=  layer.multiplier.Power*layer.n_moltiplicazioni
                #print("LayerPower: ", layer.power, "n_moltiplicazioni: ", layer.n_moltiplicazioni, "multiplier_power: ", layer.multiplier.Power )

            
        return total_power
    
    
    def return_power_for_fibers(self):
        total_power = 0
        
        for id_layer in range(len(self.layers)):
            if isinstance(self.layers[id_layer], (ConvLayer)):
                
                #! DEVO CAPIRE COME SI CALCOLA LA POTENZA PER FIBRE
                print("LEN: ", np.shape((self.layers[id_layer].mul_per_fibers)))
                print("Mul_Per_fibers_", self.layers[id_layer].mul_per_fibers)
                print("NMoltipl_",self.layers[id_layer].n_moltiplicazioni)
                exit()
                temp_power = self.layers[id_layer].mul_per_fibers * self.layers[id_layer].n_moltiplicazioni
                total_power += temp_power
                
            elif isinstance(self.layers[id_layer], (DenseLayer)):
                
                
                
                self.layers[id_layer].multiplier = dense_multiplier[dense_counter]
            #     #TODO: aggiungere update_bias_for_mul per ogni layer conv e dense
            #     #QUesto valore è un vector di bias per per ogni moltiplier
            #     #self.layers[id_layer].update_bias = self.layers[id_layer].update_bias_for_mul[idx]
                self.layers[id_layer].M = dense_multiplier[dense_counter].multiplier
                dense_counter +=1
        
        
        
        
        return total_power
    

    def save_occorenze(self,base_name):
        for layer in self.layers:
            if isinstance(layer, (ConvLayer, DenseLayer)):
                np.save(base_name+layer.name,layer.occorenza.copy_to_host())



"""
Copyright 2022  Salvatore Barone <salvatore.barone@unina.it>
                Filippo Ferrandino <fi.ferrandino@studenti.unina.it>

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3 of the License, or any later version.

This is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
RMEncoder; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
import numpy as np
from numba import cuda
from inspectnn.BaseLayer import BaseLayer
from inspectnn.Add.Add_kernel_tflite import add3d_tflite,add3dv2_tflite

from scipy.special import softmax

class SoftmaxLayer(BaseLayer):
    def __init__(self, name = "cocatenate",activation = ""):
        super().__init__(self, name = name)
        self.activation = activation

    def __deepcopy__(self, memo = None):
        return SoftmaxLayer(name = self.name)

    def load_weights(self, **kwargs):
        self.input_shape= kwargs["input_shape"]
        self.output_shape = kwargs["output_shape"]
        self.outputv = np.zeros(self.output_shape)
        self.output_mul= 1/kwargs["quant_output_k"]
        self.output_offset= kwargs["quant_output_offset"]

    def forward_pass(self, inputv):
        self.outputv = softmax(inputv/self.pre_layer.output_mul + self.pre_layer.output_offset) #TODO: controllare se è la stessa softmax di tfLITE 
            

        


import time,sys,imp,os
#import pandas 
from numba import cuda, jit
import numpy as np
import re
from inspectnn.Conv.ConvLayer import ConvLayer
from inspectnn.Dense.DenseLayer import DenseLayer
from inspectnn.Add.AddLayer import AddLayer
from inspectnn.Concatenate.ConcatenateLayer import Concatenate
class Multiply():
    def __init__(self,m,Area,Power,name="Exact"):
        self.multiplier = m
        self.Power = Power/65536.0#per moltiplicazione
        self.Area  = Area
        self.name = name

class BaseModel():
    def __init__(self, multiplier,quant_nbits = 8,input_shape=[28, 28, 1],save_csv_path=''):
        
        self.quant_nbits = quant_nbits
        self.input_shape = input_shape
        self.elapsed_time = 0
        self.total_elapsed_time=0
        self.save_csv_path=save_csv_path
        self.images_on_gpu = False

        self.layer_apx = None
        
        if None != multiplier:
            self.exact_multiplier = multiplier
        else:#Moltiplicatore di base
            dim = 128
           
            self.def_multiplier_8b = np.zeros((2*dim,2*dim),dtype=np.int16)
            self.def_multiplier_7b = np.zeros((2*dim,2*dim),dtype=np.int16)
            self.def_multiplier_6b = np.zeros((2*dim,2*dim),dtype=np.int16)
            self.def_multiplier_5b = np.zeros((2*dim,2*dim),dtype=np.int16)
            self.def_multiplier_4b = np.zeros((2*dim,2*dim),dtype=np.int16)
            self.def_multiplier_3b = np.zeros((2*dim,2*dim),dtype=np.int16)
            self.def_multiplier_2b = np.zeros((2*dim,2*dim),dtype=np.int16)
            for i in range(-dim,dim):
                for j in range(-dim,dim):
                    self.def_multiplier_8b[i+dim,j+dim]=i*j
                    self.def_multiplier_7b[i+dim,j+dim]=i*j & 0xFFFD  
                    self.def_multiplier_6b[i+dim,j+dim]=i*j & 0xFFFB  
                    self.def_multiplier_5b[i+dim,j+dim]=i*j & 0xFFF9  
                    self.def_multiplier_4b[i+dim,j+dim]=(i*j) & 0xFFF7  
                    self.def_multiplier_3b[i+dim,j+dim]=(i*j) & 0xFFF5  
                    self.def_multiplier_2b[i+dim,j+dim]=(i*j) & 0xFFF3  
            
            self.exact_multiplier = Multiply(cuda.to_device(self.def_multiplier_8b),729.8,0.425)
            
            self.def_multiplier_8b = Multiply(cuda.to_device(self.def_multiplier_8b),800,0.410,"8b")
            self.def_multiplier_7b = Multiply(cuda.to_device(self.def_multiplier_7b),600,0.300,"7b")
            self.def_multiplier_6b = Multiply(cuda.to_device(self.def_multiplier_6b),400,0.200,"6b")
            self.def_multiplier_5b = Multiply(cuda.to_device(self.def_multiplier_5b),350,0.125,"5b")
            self.def_multiplier_4b = Multiply(cuda.to_device(self.def_multiplier_4b),200,0.100,"4b")
            self.def_multiplier_3b = Multiply(cuda.to_device(self.def_multiplier_3b),100,0.050,"3b")
            self.def_multiplier_2b = Multiply(cuda.to_device(self.def_multiplier_2b),50 ,0.025,"2b")

        self.all_multiplier_quant = [self.def_multiplier_2b,self.def_multiplier_3b,self.def_multiplier_4b,self.def_multiplier_5b,self.def_multiplier_6b,self.def_multiplier_7b,self.def_multiplier_8b]
        self.net = None

   
    def evaluate(self,images,labels,log=False,reset_statistics=False):
        if reset_statistics:
            for id_layer in range(len(self.net.layers)):
                self.net.elapsed_time[id_layer] = 0
                
        
        num_of_samples=len(images)
        if log : 
            print(f"Testing on {len(images)} images.")
        st = time.time()
        if self.images_on_gpu:
            Accuracy = self.net.evaluate_accuracy(self.labels, self.images)
        else:
            Accuracy = self.net.evaluate_accuracy(labels, images)
        if log :
            print(f"Accuracy: {Accuracy}")
        et = time.time()
        self.elapsed_time = et - st
        self.total_elapsed_time+=self.elapsed_time
        
        
        if log :
            print('MIMT execution time:', self.elapsed_time, 'seconds')
            print('FPS:', num_of_samples/self.elapsed_time) 
            self.net.print_time_statics()
        return Accuracy
    
    def evaluate_pure(self):

        st = time.time()
        
        #Accuracy = self.net.evaluate_accuracy(self.labels, self.images)
        Accuracy = self.net.evaluate_accuracy(self.labels, self.images) #TODO ulteriori esperimenti da fare, da sostituire con la versione avanzata con la sensitivita

        et = time.time()
        self.elapsed_time = et - st
        self.total_elapsed_time+=self.elapsed_time

        return Accuracy
    
    def evaluate_pure_AdversialAttack(self):

        st = time.time()
        
        #Accuracy = self.net.evaluate_accuracy(self.labels, self.images)
        Accuracy = self.net.evaluate_accuracy(self.AdversialAttack.labels, self.AdversialAttack.images) #TODO ulteriori esperimenti da fare, da sostituire con la versione avanzata con la sensitivita

        et = time.time()
        self.elapsed_time = et - st
        self.total_elapsed_time+=self.elapsed_time

        return Accuracy
    
    def evaluate_baseline_accuracy(self,images_apx,labels_apx,reset_statistics=False):
        self.net.update_multipler([self.exact_multiplier])
    
        self.baseline_accuracy = self.evaluate(images_apx, labels_apx,reset_statistics=reset_statistics)

    def evaluate_all(self,images_apx,labels_apx,print_results=False):
        results = []
        id = 0
        for multiplier in self.all_multiplier:
   
            self.net.update_multipler([multiplier])
            accuracy = self.evaluate(images_apx, labels_apx)
            Power = self.net.return_Power()
            results.append([multiplier.name, Power,accuracy, self.baseline_accuracy - accuracy, self.elapsed_time])
            if print_results:
                print(f"{id}/{len(self.all_multiplier):}" ,results[-1])

            id+=1
        return results
    
    def load_all_multiply_for_layer(self,configuration):
        self.all_multiplier = []
        net = self.net
        
        idx=0
        vector_list_mul_data = configuration["axmuls"]
        
        data_model = None
        for  list_mul_data in vector_list_mul_data:

            for mul_data in list_mul_data:
                multiplier, _ = self._load_multiplier(mul_data["path"])
                multiplier.name=mul_data["name"]
                
                multiplier.MAE   =mul_data['MAE']
                multiplier.AWCE  =mul_data['AWCE']
                multiplier.MRE   =mul_data['MRE']
                multiplier.Power =mul_data['power']/65536.0#per moltiplicazione
                multiplier.Area  =mul_data['area']
    
                if data_model is None:
                    data_model = [multiplier.model]
                else:
                    data_model = np.concatenate((data_model,[multiplier.model]))

                self.all_multiplier.append(multiplier)
                
        
        self.all_multiplier_gpu = cuda.to_device(np.array(data_model,dtype=np.int16))

        self.net.load_total_multiply_gpu(self.all_multiplier_gpu)

        return self.all_multiplier, data_model

    def load_all_multiply(self, config, filter=None):
        self.all_multiplier = []

        for axmul in config["ax_muls"]:
            multiplier_name = axmul["name"]
            multiplier_file = f"mul8s_{multiplier_name}.py"
            found_multiplier = False

            for root, dirs, files in os.walk(config["path"]):
                for f in files:
                    if f == multiplier_file and ((filter is not None and filter in f) or filter is None):
                        py_filename = os.path.join(root, f)
                        multiplier, _ = self._load_multiplier(py_filename)
                        multiplier.filename = py_filename.split('/')[-2]
                        multiplier.MAE = axmul['MAE']
                        multiplier.AWCE = axmul['AWCE']
                        multiplier.MRE = axmul['MRE']
                        multiplier.Power = axmul['power'] / 65536.0  # per moltiplicazione
                        multiplier.Area = axmul['area']

                        self.all_multiplier.append(multiplier)
                        found_multiplier = True
                        print(f"Loaded multiplier: {multiplier_name}, object {multiplier}")
                        break  # Stop searching for the same multiplier name in other directories

            if not found_multiplier:
                print(f"Multiplier file not found for {multiplier_name}.")
                # Handle the error or raise an exception if needed
        
        return self.all_multiplier
    
    
    
    def load_all_multiply2(self,approxmults_path,config_file,filter=None):
        self.all_multiplier = []
        
        data_model = None
        
        for root, dirs, files in os.walk(approxmults_path):
            dirs[:] = [d for d in dirs if not (d[0] == "." or d[0] == "_")]
            for f in files:
                if (
                    f.endswith('.py') and 
                    ((filter is not None and filter in f) or filter is None) and
                    f[0] != "." and f[0] != "_"
                ):
                    py_filename = os.path.join(root, f)
                    multiplier, _ = self._load_multiplier(py_filename)
                    multiplier.filename=py_filename.split('/')[-2]                 
                    
                    self.all_multiplier.append(multiplier)
                    
        data_model = None 
        if data_model is None:
            data_model = [multiplier.model]
        else:
            data_model = np.concatenate((data_model,[multiplier.model]))

        self.all_multiplier.append(multiplier)

        self.all_multiplier_gpu = cuda.to_device(np.array(data_model,dtype=np.int16))
        
        return self.all_multiplier
                   
    
    def _import_module(self,module_name):
        # Fast path: see if the module has already been imported.
        try:
            return sys.modules[module_name]
        except KeyError:
            pass
        # If any of the following calls raises an exception,
        # there's a problem we can't handle -- let the caller handle it.
        fp, pathname, description = imp.find_module(module_name)
        try:
            return imp.load_module(module_name, fp, pathname, description)    
        finally:
            # Since we may exit via an exception, close fp explicitly.
            if fp:
                fp.close()


    def _load_multiplier(self,filename):
        module_name = os.path.basename(filename)
        name_class = os.path.splitext(module_name)[0]
        name_import = os.path.splitext(filename)[0]
        
        module_name = self._import_module(name_import)
        variant_mul = getattr(module_name, name_class)
        mult_class = variant_mul()
        
        mult_class.multiplier = cuda.to_device(np.array(mult_class.model,dtype=np.int16))
        return mult_class, name_class
    

    def load_img_gpu(self,images,labels):
        self.images_on_gpu=True
        self.labels = labels
        self.images = []
        for i in range(len(images)):
            self.images.append(cuda.to_device(np.ascontiguousarray(images[i], dtype=np.int16)))
            
            
    def load_AdversialAttack_img_gpu(self,images,labels):
        class AdversialAttack():
            def __init__(self,images,labels):
                self.labels = [] 
                self.images = []
                for i in range(len(images)):
                    self.images.append(cuda.to_device(np.ascontiguousarray(images[i], dtype=np.int32)))
                    self.labels.append(labels[i])
                    
        self.AdversialAttack = AdversialAttack(images,labels)
       

    def load_img_quantizate_on_gpu(self,images,labels):
        self.images_on_gpu=True
        self.labels = labels
        self.images = []
        
        for i in range(len(images)):
            self.net.layers[0].forward_pass(inputv = images[i])
            self.images.append(cuda.to_device(np.ascontiguousarray(self.net.layers[0].results.copy_to_host(), dtype=np.int8)))

        self.net.init_quantizate = False

    def generate_multipler_list(self,id,mul=None):
        #print(id)
        if mul is None:
            mul = self.all_multiplier
            #mul = self.all_multiplier_quant
        multipler_list= []
        for i in id:
            multipler_list.append(mul[i])
        return multipler_list